from abc import ABCMeta, abstractmethod
from typing import List, Generic, TypeVar, Type, Optional

import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)

logger = structlog.get_logger(__file__)


# Define the TypeVar for the generic type
T = TypeVar("T", bound=BaseVendorApiResponse)


class StandardizingVendor(Generic[T], metaclass=ABCMeta):
    """
    for now, we just save the version attached to the vendor response, in future we need to expand it
    to have different parsers for possible new versions
    """

    def __init__(self, version: str = "1") -> None:
        self._version: str = version

    @abstractmethod
    async def standardize_async(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse[T]]:
        """
        returns the vendor specific response from the vendor
        """

    @abstractmethod
    def vendor_specific_to_std(
        self,
        *,
        vendor_specific_addresses: List[VendorResponse[T]],
        raw_addresses: List[RawAddress],
    ) -> List[StandardizedAddress]:
        """
        each vendor class knows how to convert its response to StdAddress

        :param vendor_specific_addresses: vendor specific addresses
        :param raw_addresses: raw addresses
        :return: standardized addresses
        """

    @abstractmethod
    def batch_request_max_size(self) -> int:
        return 0

    @classmethod
    @abstractmethod
    def get_vendor_name(cls) -> str:
        """
        returns the name of the vendor
        """
        return ""

    def get_version(self) -> str:
        return self._version

    @abstractmethod
    def _to_vendor_response(
        self,
        vendor_response: List[T],
        raw_addresses: List[RawAddress],
        vendor_name: str,
        response_version: str,
    ) -> List[VendorResponse[T]]:
        """
        using RecordID, we want to find the corresponding raw address of a vendor response
        """

    @classmethod
    @abstractmethod
    def get_api_response_class(cls) -> Type[T]:
        pass

    # noinspection PyMethodMayBeStatic
    def get_matching_address(
        self, raw_addresses: List[RawAddress], address_id: str
    ) -> Optional[RawAddress]:
        """
        get the matching address from the list of raw addresses

        :param raw_addresses: list of raw addresses
        :param address_id: address id
        :return: matching address if found
        """
        return next(iter([r for r in raw_addresses if r.get_id() == address_id]))



from typing import Dict, List

import structlog

from spark_pipeline_framework.utilities.helix_geolocation.v1.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v1.vendor_response import (
    VendorResponse,
)

logger = structlog.get_logger(__file__)


class GeocodioStandardizingVendor(StandardizingVendor):
    def standardize(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse]:
        vendor_specific_addresses: List[Dict[str, str]] = []
        for address in raw_addresses:
            address_dict = address.to_dict()

            vendor_specific_addresses.append(address_dict)

        vendor_responses: List[VendorResponse] = super()._to_vendor_response(
            vendor_response=vendor_specific_addresses,
            raw_addresses=raw_addresses,
            vendor_name=self.get_vendor_name(),
            response_version=self.get_version(),
        )
        return vendor_responses

    @staticmethod
    def vendor_specific_to_std(
        vendor_specific_addresses: List[VendorResponse],
    ) -> List[StandardizedAddress]:
        """
        each vendor class knows how to convert its response to StdAddress
        """
        std_addresses = [
            StandardizedAddress(
                address_id=a.api_call_response["RecordID"],
                line1=a.api_call_response["line1"],
                line2=a.api_call_response["line2"],
                city=a.api_call_response["city"],
                zipcode=a.api_call_response["zipcode"],
                state=a.api_call_response["state"],
                country=a.api_call_response["geocodio_Country"],
                latitude=a.api_call_response["Latitude"],
                longitude=a.api_call_response["Longitude"],
                standardize_vendor="Geocodio",
            )
            for a in vendor_specific_addresses
            if any(a)
        ]
        return std_addresses

    @staticmethod
    def get_vendor_name() -> str:
        return "Geocodio"

    @staticmethod
    def batch_request_max_size() -> int:
        return 100



class ElasticSearchSenderException(Exception):
    def __init__(
        self, exception: Exception, url: str, json_data: str, message: str
    ) -> None:
        self.exception: Exception = exception
        self.url: str = url
        self.data: str = json_data
        super().__init__(f"ElasticSearch send failed to {url}: {json_data}.  {message}")



from json import JSONDecodeError
from pathlib import Path
from typing import List, Type

from pyspark.sql import DataFrame, SparkSession

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_exception_handler.v1.framework_exception_handler import (
    FrameworkExceptionHandlerTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_framework_exception_handler(spark_session: SparkSession) -> None:
    # create a dataframe with the test data
    data_dir: Path = Path(__file__).parent.joinpath("./")
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    invalid_view: str = "invalid_view"
    valid_view: str = "valid_view"

    with ProgressLogger() as progress_logger:
        FrameworkExceptionHandlerTransformer(
            name="Exception Handler Test",
            stages=[
                # A step that tries to load a non-existent CSV file (should fail)
                FrameworkCsvLoader(
                    view=valid_view,
                    file_path=data_dir.joinpath("invalid_location.csv"),
                    clean_column_names=False,
                )
            ],
            exception_stages=[
                FrameworkCsvLoader(
                    view=invalid_view,
                    file_path=data_dir.joinpath("primary_care_protocol.csv"),
                    clean_column_names=False,
                )
            ],
            retry_count=2,
            raise_on_exception=False,
            progress_logger=progress_logger,
        ).transform(df)
    result_df: DataFrame = spark_session.table(invalid_view)

    # Assert that the `exception_stages` has successfully run and data is loaded
    assert result_df.count() == 3
    # Verify that the invalid view was NOT created, confirming that the original `stages` failed
    assert not spark_session.catalog.tableExists(valid_view)


def test_framework_exception_handler_with_unhandled_exception(
    spark_session: SparkSession,
) -> None:
    """Test that unhandled exception types are NOT processed by exception stages"""
    data_dir: Path = Path(__file__).parent.joinpath("./")
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    invalid_view: str = "unhandled_invalid_view"
    valid_view: str = "unhandled_valid_view"

    # Only handle JSONDecodeError, but the CSV loader will raise AnalysisException due to PATH_NOT_FOUND
    handled_exceptions: List[Type[BaseException]] = [JSONDecodeError]

    with ProgressLogger() as progress_logger:
        try:
            FrameworkExceptionHandlerTransformer(
                name="Unhandled Exception Test",
                stages=[
                    FrameworkCsvLoader(
                        view=invalid_view,
                        file_path=data_dir.joinpath("non_existent_file.csv"),
                        clean_column_names=False,
                    )
                ],
                exception_stages=[
                    FrameworkCsvLoader(
                        view=valid_view,
                        file_path=data_dir.joinpath("primary_care_protocol.csv"),
                        clean_column_names=False,
                    )
                ],
                handled_exceptions=handled_exceptions,
                raise_on_exception=True,  # Should re-raise the unhandled exception
                progress_logger=progress_logger,
            ).transform(df)

        except Exception as e:
            # Verify that the exception was NOT handled and was not a JSONDecodeError
            assert not isinstance(e, JSONDecodeError), "Should not be a JSONDecodeError"

    # Verify that the `exception_stages` did NOT run
    assert not spark_session.catalog.tableExists(valid_view)

    # Verify that the `stages` also failed
    assert not spark_session.catalog.tableExists(invalid_view)



import pytest
import asyncio
from typing import AsyncGenerator, Dict
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType, StructField, StringType
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper


# Sample async generator for testing
async def sample_async_generator() -> AsyncGenerator[int, None]:
    for i in range(5):
        yield i


async def sample_data_frame_async_generator() -> AsyncGenerator[Dict[str, str], None]:
    for i in range(5):
        yield {"value": str(i)}


@pytest.fixture
def sample_schema() -> StructType:
    return StructType([StructField("value", StringType(), True)])


@pytest.mark.asyncio
async def test_collect_items() -> None:
    generator = sample_async_generator()
    result = await AsyncHelper.collect_items(generator)
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_collect_async_data() -> None:
    generator = sample_async_generator()
    result = [
        chunk
        async for chunk in AsyncHelper.collect_async_data(
            async_gen=generator, chunk_size=2
        )
    ]
    assert result == [[0, 1], [2, 3], [4]]


@pytest.mark.asyncio
async def test_async_generator_to_dataframe(
    spark_session: SparkSession, sample_schema: StructType
) -> None:
    generator = sample_data_frame_async_generator()
    df = spark_session.createDataFrame([], sample_schema)
    result_df = await AsyncHelper.async_generator_to_dataframe(
        df, generator, sample_schema, results_per_batch=2
    )
    result = result_df.collect()
    assert len(result) == 5
    assert result[0]["value"] == "0"
    assert result[1]["value"] == "1"


def test_run() -> None:
    async def sample_coroutine() -> int:
        await asyncio.sleep(0.1)
        return 42

    result = AsyncHelper.run(sample_coroutine())
    assert result == 42


def test_run_in_new_thread_and_wait() -> None:
    async def sample_coroutine() -> int:
        await asyncio.sleep(0.1)
        return 42

    result = AsyncHelper.run_in_new_thread_and_wait(sample_coroutine())
    assert result == 42


def test_run_in_thread_pool_and_wait() -> None:
    async def sample_coroutine() -> int:
        await asyncio.sleep(0.1)
        return 42

    result = AsyncHelper.run_in_thread_pool_and_wait(sample_coroutine())
    assert result == 42



from dataclasses import dataclass
from typing import Optional

from spark_pipeline_framework.transformers.http_data_receiver.v5.common import (
    RESPONSE_PROCESSOR_TYPE,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v3.oauth2_client_credentials_flow import (
    OAuth2Credentails,
)


@dataclass
class HttpDataReceiverParameters:
    log_level: str
    response_processor: RESPONSE_PROCESSOR_TYPE
    raise_error: bool
    credentials: OAuth2Credentails | None
    auth_url: Optional[str]
    cert: str | tuple[str, str] | None
    verify: bool | str | None
    pandas_udf_parameters: AsyncPandasUdfParameters



import dataclasses
from typing import Any, Dict, Optional

# noinspection PyPep8Naming
from pyspark.sql.types import (
    StringType,
    StructType,
)
from pyspark.sql.types import StructField


@dataclasses.dataclass
class GetBatchError:
    url: str
    status_code: int
    error_text: str
    request_id: Optional[str]

    @staticmethod
    def get_schema() -> StructType:
        return StructType(
            [
                StructField("url", StringType(), True),
                StructField("status_code", StringType(), True),
                StructField("error_text", StringType(), True),
                StructField("request_id", StringType(), True),
            ]
        )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)



from typing import Dict, Any, Optional

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import (
    struct,
    lit,
    col,
    concat,
    array,
    expr,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkFhirMetaUpdater(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        resource_type: str,
        source_view: str,
        view: str,
        source_url: str,
        owner_code: str,
        access_code: Optional[str] = None,
        vendor_code: Optional[str] = None,
        source_assigning_authority_code: Optional[str] = None,
        connection_type_code: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Used to add/update the meta field with the b.well mandatory stuff
        :param resource_type: type of resource to update
        :param source_view: read from this view
        :param view: write to this view
        :param source_url: URL for `meta.source`
        :param owner_code: owner security tag
        :param access_code: access security tag
        :param vendor_code: vendor security tag
        :param source_assigning_authority_code: sourceAssigningAuthority security tag
        :param connection_type_code: connection_type security tag
        :param name: Name for the transformer
        :param parameters: Parameters
        :param progress_logger: ProgressLogger
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.resource_type: Param[str] = Param(self, "resource_type", "")
        self._setDefault(resource_type=None)

        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=None)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.source_url: Param[str] = Param(self, "source_url", "")
        self._setDefault(source_url=None)

        self.owner_code: Param[str] = Param(self, "owner_code", "")
        self._setDefault(owner_code=None)

        self.access_code: Param[Optional[str]] = Param(self, "access_code", "")
        self._setDefault(access_code=access_code)

        self.vendor_code: Param[Optional[str]] = Param(self, "vendor_code", "")
        self._setDefault(vendor_code=vendor_code)

        self.source_assigning_authority_code: Param[Optional[str]] = Param(
            self, "source_assigning_authority_code", ""
        )
        self._setDefault(
            source_assigning_authority_code=source_assigning_authority_code
        )

        self.connection_type_code: Param[Optional[str]] = Param(
            self, "connection_type_code", ""
        )
        self._setDefault(connection_type_code=connection_type_code)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        resource_type: str = self.getOrDefault(self.resource_type)
        source_view: str = self.getOrDefault(self.source_view)
        view: str = self.getOrDefault(self.view)
        source_url: str = self.getOrDefault(self.source_url)
        owner_code: str = self.getOrDefault(self.owner_code)
        access_code: Optional[str] = self.getOrDefault(self.access_code)
        vendor_code: Optional[str] = self.getOrDefault(self.vendor_code)
        source_assigning_authority_code: Optional[str] = self.getOrDefault(
            self.source_assigning_authority_code
        )
        connection_type_code: Optional[str] = self.getOrDefault(
            self.connection_type_code
        )

        df = df.sparkSession.table(source_view)

        owner_codeset: str = "https://www.icanbwell.com/owner"
        access_codeset: str = "https://www.icanbwell.com/access"
        vendor_codeset: str = "https://www.icanbwell.com/vendor"
        source_assigning_authority_codeset: str = (
            "https://www.icanbwell.com/sourceAssigningAuthority"
        )
        connection_type_codeset: str = "https://www.icanbwell.com/connectionType"

        df = df.withColumn(
            "meta",
            struct(
                concat(lit(source_url), lit(f"/{resource_type}/"), col("id")).alias(
                    "source"
                ),
                array(
                    struct(
                        lit(owner_codeset).alias("system"),
                        lit(owner_code).alias("code"),
                    ),
                    struct(
                        lit(access_codeset).alias("system"),
                        lit(access_code).alias("code"),
                    ),
                    struct(
                        lit(vendor_codeset).alias("system"),
                        lit(vendor_code).alias("code"),
                    ),
                    struct(
                        lit(source_assigning_authority_codeset).alias("system"),
                        lit(source_assigning_authority_code).alias("code"),
                    ),
                    struct(
                        lit(connection_type_codeset).alias("system"),
                        lit(connection_type_code).alias("code"),
                    ),
                ).alias("security"),
            ),
        )
        df = df.withColumn(
            "meta",
            expr(
                "named_struct('source', meta.source, 'security', filter(meta.security, s -> s.code IS NOT NULL))"
            ),
        )

        df.createOrReplaceTempView(view)
        return df



import json
from os import environ
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Callable

from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from helix_fhir_client_sdk.responses.fhir_update_response import FhirUpdateResponse
from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from pyspark import StorageLevel
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, get_json_object, to_json, struct
from pyspark.sql.types import Row
from pyspark.errors import AnalysisException, PythonException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender_parameters import (
    FhirSenderParameters,
)
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender_processor import (
    FhirSenderProcessor,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token_async,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_item import (
    FhirMergeResponseItem,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_item_schema import (
    FhirMergeResponseItemSchema,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_validation_exception import (
    FhirSenderValidationException,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.map_functions import remove_field_from_json
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class FhirSender(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        server_url: str,
        file_path: Path | str | Callable[[str | None], Path | str] | None = None,
        response_path: Path | str | Callable[[str | None], Path | str] | None = None,
        resource: str,
        batch_size: Optional[int] = 30,
        limit: int = -1,
        file_format: str = "json",
        validation_server_url: Optional[str] = None,
        throw_exception_on_validation_failure: Optional[bool] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        auth_well_known_url: Optional[str] = None,
        operation: Union[
            FhirSenderOperation, str
        ] = FhirSenderOperation.FHIR_OPERATION_MERGE.value,
        name: Optional[str] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        error_view: Optional[str] = None,
        view: Optional[str] = None,
        ignore_status_codes: Optional[List[int]] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        delta_lake_table: Optional[str] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        run_synchronously: Optional[bool] = None,
        sort_by_column_name_and_type: Optional[tuple[str, Any]] = None,
        drop_fields_from_json: Optional[List[str]] = None,
        source_view: Optional[str] = None,
        log_level: Optional[str] = None,
        max_chunk_size: int = 100,
        process_chunks_in_parallel: Optional[bool] = True,
        maximum_concurrent_tasks: int = 100,
        smart_merge: Optional[bool] = None,
    ):
        """
        Sends FHIR json stored in a folder to a FHIR server

        :param server_url: url for the fhir server
        :param file_path: where to read the json from
        :param response_path: where to store the responses
        :param resource: name of resource to send to fhir server
        :param batch_size: how many resources to send in one batch call to fhir server
        :param validation_server_url: (Optional) url to use for validating FHIR resources
        :param throw_exception_on_validation_failure: whether to continue after a validation failure
        :param auth_server_url: url to auth server
        :param auth_client_id: client id
        :param auth_client_secret: client secret
        :param auth_login_token: login token to send auth server
        :param auth_scopes: scopes to request
        :param additional_request_headers: (Optional) Additional request headers to use
                                            (Eg: {"Accept-Charset": "utf-8"})
        :param operation: What FHIR operation to perform (e.g. $merge, delete, etc.)
        :param mode: if output files exist, should we overwrite or append
        :param error_view: (Optional) log errors into this view (view only exists IF there are errors)
                            and don't throw exceptions.
                            schema: id, resourceType, issue
        :param view: (Optional) store merge result in this view
        :param delta_lake_table: use delta lake format
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param run_synchronously: (Optional) Run on the Spark master to make debugging easier on dev machines
        :param sort_by_column_name_and_type: (Optional) tuple of columnName, columnType to be used for sorting
        :param drop_fields_from_json: (Optional) List of field names to drop from json
        :param source_view: (Optional) source view to read from
        :param log_level: (Optional) log level
        :param max_chunk_size: (Optional) max chunk size
        :param process_chunks_in_parallel: (Optional) process chunks in parallel
        :param maximum_concurrent_tasks: (Optional) maximum concurrent tasks
        :param smart_merge: (Optional) whether to use smart merge or not
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert file_path or source_view, "Either file_path or source_view must be set"
        if file_path is not None:
            assert (
                isinstance(file_path, Path)
                or isinstance(file_path, str)
                or callable(file_path)
            ), type(file_path)
        else:
            assert source_view, "source_view must be set"

        assert response_path or view, "Either response_path or view must be set"
        if response_path is not None:
            assert (
                isinstance(response_path, Path)
                or isinstance(response_path, str)
                or callable(response_path)
            ), type(response_path)
        else:
            assert view, "view must be set"

        assert progress_logger

        assert server_url

        self.log_level: Param[str] = Param(self, "log_level", "")
        self._setDefault(log_level=log_level)

        self.logger = get_logger(__name__, level=log_level or "INFO")

        self.server_url: Param[str] = Param(self, "server_url", "")
        self._setDefault(server_url=None)

        self.file_path: Param[
            Path | str | Callable[[str | None], Path | str] | None
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.response_path: Param[
            Path | str | Callable[[str | None], Path | str] | None
        ] = Param(self, "response_path", "")
        self._setDefault(response_path=None)

        self.resource: Param[str] = Param(self, "resource", "")
        self._setDefault(resource=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.throw_exception_on_validation_failure: Param[Optional[bool]] = Param(
            self, "throw_exception_on_validation_failure", ""
        )
        self._setDefault(
            throw_exception_on_validation_failure=throw_exception_on_validation_failure
        )

        self.file_format: Param[str] = Param(self, "file_format", "")
        self._setDefault(file_format=file_format)

        self.validation_server_url: Param[Optional[str]] = Param(
            self, "validation_server_url", ""
        )
        self._setDefault(validation_server_url=None)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[List[str]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.auth_well_known_url: Param[Optional[str]] = Param(
            self, "auth_well_known_url", ""
        )
        self._setDefault(auth_well_known_url=None)

        self.operation: Param[Union[FhirSenderOperation, str]] = Param(
            self, "operation", ""
        )
        self._setDefault(operation=operation)

        self.additional_request_headers: Param[Optional[Dict[str, str]]] = Param(
            self, "additional_request_headers", ""
        )
        self._setDefault(additional_request_headers=additional_request_headers)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.retry_count: Param[Optional[int]] = Param(self, "retry_count", "")
        self._setDefault(retry_count=None)

        self.exclude_status_codes_from_retry: Param[Optional[List[int]]] = Param(
            self, "exclude_status_codes_from_retry", ""
        )
        self._setDefault(exclude_status_codes_from_retry=None)

        self.ignore_status_codes: Param[Optional[List[int]]] = Param(
            self, "ignore_status_codes", ""
        )
        self._setDefault(ignore_status_codes=None)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.sort_by_column_name_and_type: Param[Optional[tuple[str, Any]]] = Param(
            self, "sort_by_column_name_and_type", ""
        )
        self._setDefault(sort_by_column_name_and_type=sort_by_column_name_and_type)

        self.drop_fields_from_json: Param[Optional[List[str]]] = Param(
            self, "drop_fields_from_json", ""
        )
        self._setDefault(drop_fields_from_json=drop_fields_from_json)

        self.source_view: Param[Optional[str]] = Param(self, "source_view", "")
        self._setDefault(source_view=None)

        self.max_chunk_size: Param[int] = Param(self, "max_chunk_size", "")
        self._setDefault(max_chunk_size=max_chunk_size)

        self.process_chunks_in_parallel: Param[Optional[bool]] = Param(
            self, "process_chunks_in_parallel", ""
        )
        self._setDefault(process_chunks_in_parallel=process_chunks_in_parallel)

        self.maximum_concurrent_tasks: Param[int] = Param(
            self, "maximum_concurrent_tasks", ""
        )
        self._setDefault(maximum_concurrent_tasks=maximum_concurrent_tasks)

        self.smart_merge: Param[Optional[bool]] = Param(self, "smart_merge", "")
        self._setDefault(smart_merge=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        file_path: Path | str | Callable[[str | None], Path | str] | None = (
            self.getOrDefault(self.file_path)
        )
        if file_path is not None and callable(file_path):
            file_path = file_path(self.loop_id)
        response_path: Path | str | Callable[[str | None], Path | str] | None = (
            self.getOrDefault(self.response_path)
        )
        if response_path is not None and callable(response_path):
            response_path = response_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        resource_name: str = self.getOrDefault(self.resource)
        parameters = self.getParameters()
        additional_request_headers: Optional[Dict[str, str]] = self.getOrDefault(
            self.additional_request_headers
        )
        server_url: str = self.getOrDefault(self.server_url)
        batch_size: Optional[int] = self.getOrDefault(self.batch_size)
        throw_exception_on_validation_failure: Optional[bool] = self.getOrDefault(
            self.throw_exception_on_validation_failure
        )
        operation: Union[FhirSenderOperation, str] = self.getOrDefault(self.operation)
        mode: str = self.getOrDefault(self.mode)

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )
        sort_by_column_name_and_type: Optional[tuple[str, Any]] = self.getOrDefault(
            self.sort_by_column_name_and_type
        )
        drop_fields_from_json: Optional[List[str]] = self.getOrDefault(
            self.drop_fields_from_json
        )

        if not batch_size or batch_size == 0:
            batch_size = 30
        else:
            batch_size = int(batch_size)

        file_format: str = self.getOrDefault(self.file_format)
        validation_server_url: Optional[str] = self.getOrDefault(
            self.validation_server_url
        )
        auth_server_url: Optional[str] = self.getOrDefault(self.auth_server_url)
        auth_well_known_url: Optional[str] = self.getOrDefault(self.auth_well_known_url)
        auth_client_id: Optional[str] = self.getOrDefault(self.auth_client_id)
        auth_client_secret: Optional[str] = self.getOrDefault(self.auth_client_secret)
        auth_login_token: Optional[str] = self.getOrDefault(self.auth_login_token)
        auth_scopes: Optional[List[str]] = self.getOrDefault(self.auth_scopes)
        auth_access_token: Optional[str] = None

        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view: Optional[str] = self.getOrDefault(self.view)

        log_level: Optional[str] = self.getOrDefault(self.log_level) or environ.get(
            "LOGLEVEL"
        )

        ignore_status_codes: List[int] | None = (
            self.getOrDefault(self.ignore_status_codes) or []
        )
        if ignore_status_codes is None:
            ignore_status_codes = []
        ignore_status_codes.append(200)

        retry_count: Optional[int] = self.getOrDefault(self.retry_count)
        exclude_status_codes_from_retry: Optional[List[int]] = self.getOrDefault(
            self.exclude_status_codes_from_retry
        )

        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)

        source_view: Optional[str] = self.getOrDefault(self.source_view)

        max_chunk_size: int = self.getOrDefault(self.max_chunk_size)
        process_chunks_in_parallel: Optional[bool] = self.getOrDefault(
            self.process_chunks_in_parallel
        )
        maximum_concurrent_tasks: int = self.getOrDefault(self.maximum_concurrent_tasks)

        smart_merge: Optional[bool] = self.getOrDefault(self.smart_merge)

        if parameters and parameters.get("flow_name"):
            if additional_request_headers:
                additional_request_headers.update(
                    {
                        "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                    }
                )
            else:
                additional_request_headers = {
                    "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                }

        # get access token first so we can reuse it
        if auth_client_id:
            auth_access_token_result: GetAccessTokenResult = (
                await fhir_get_access_token_async(
                    logger=self.logger,
                    server_url=server_url,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    log_level=log_level,
                    auth_well_known_url=auth_well_known_url,
                )
            )
            auth_access_token = auth_access_token_result.access_token

        self.logger.info(
            f"Calling {server_url}/{resource_name} with client_id={auth_client_id} and scopes={auth_scopes}"
        )
        with ProgressLogMetric(
            name=f"{name}_fhir_sender", progress_logger=progress_logger
        ):
            # read all the files at path into a dataframe
            try:
                if source_view:
                    json_df: DataFrame = df.sparkSession.table(source_view)
                else:
                    path_to_files: str = str(file_path)
                    if delta_lake_table:
                        json_df = df.sparkSession.read.format("delta").load(
                            path_to_files
                        )
                    elif file_format == "parquet":
                        json_df = df.sparkSession.read.format(file_format).load(
                            path_to_files
                        )
                    else:
                        json_df = df.sparkSession.read.text(
                            path_to_files,
                            pathGlobFilter="*.json",
                            recursiveFileLookup=True,
                        )
                # if source dataframe is not a single column dataframe then convert it to a single column dataframe
                # the column name should be "value"
                # we use string column because we accept resources of different resourceTypes that have different
                # schemas
                if len(json_df.columns) > 1 or "value" not in json_df.columns:
                    json_df = json_df.select(to_json(struct("*")).alias("value"))
            except AnalysisException as e:
                if str(e).startswith("Path does not exist:"):
                    if progress_logger:
                        progress_logger.write_to_log(f"Folder {path_to_files} is empty")
                    return df
                raise

            row_count: int = json_df.count()
            if not spark_is_data_frame_empty(df=json_df):
                self.logger.info(
                    f"----- Sending {resource_name} (rows={row_count}) to FHIR server {server_url}"
                    f" with batch_size {batch_size}  -----"
                )
                assert batch_size and batch_size > 0
                # get the number of partitions
                total_partitions: int = json_df.rdd.getNumPartitions()

                if sort_by_column_name_and_type:
                    column_name, column_type = sort_by_column_name_and_type
                    json_df = json_df.withColumn(
                        column_name,
                        get_json_object(col("value"), f"$.{column_name}").cast(
                            column_type
                        ),
                    ).sortWithinPartitions(column_name)
                    json_df = json_df.drop(column_name)

                if drop_fields_from_json:
                    json_schema = json_df.schema
                    # removing sorted field from all the json values
                    json_df = json_df.rdd.map(
                        lambda row: Row(
                            value=remove_field_from_json(
                                row["value"], drop_fields_from_json
                            ),
                        ),
                    ).toDF(json_schema)

                self.logger.info(
                    f"----- Total Batches for {resource_name}: {total_partitions}  -----"
                )

                sender_parameters: FhirSenderParameters = FhirSenderParameters(
                    total_partitions=total_partitions,
                    operation=operation,
                    server_url=server_url,
                    resource_name=resource_name,
                    name=name,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    auth_access_token=auth_access_token,
                    auth_well_known_url=auth_well_known_url,
                    additional_request_headers=additional_request_headers,
                    log_level=log_level,
                    batch_size=batch_size,
                    validation_server_url=validation_server_url,
                    retry_count=retry_count,
                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                    pandas_udf_parameters=AsyncPandasUdfParameters(
                        log_level=log_level,
                        max_chunk_size=max_chunk_size,
                        process_chunks_in_parallel=process_chunks_in_parallel,
                        maximum_concurrent_tasks=maximum_concurrent_tasks,
                    ),
                    smart_merge=smart_merge,
                )
                if run_synchronously:
                    rows_to_send: List[Dict[str, Any]] = [
                        r.asDict(recursive=True) for r in json_df.collect()
                    ]
                    result_rows: List[
                        FhirMergeResponse | FhirUpdateResponse | FhirDeleteResponse
                    ] = await AsyncHelper.collect_items(
                        FhirSenderProcessor.send_partition_to_server_async(
                            partition_index=0,
                            chunk_index=0,
                            rows=rows_to_send,
                            parameters=sender_parameters,
                        )
                    )
                    merge_items: List[FhirMergeResponseItem] = (
                        FhirMergeResponseItem.from_responses(responses=result_rows)
                    )
                    result_df = (
                        df.sparkSession.createDataFrame(  # type:ignore[type-var]
                            merge_items,
                            schema=FhirMergeResponseItemSchema.get_schema(),
                        )
                    )
                else:
                    # use mapInPandas
                    # https://spark.apache.org/docs/latest/api/python/user_guide/sql/arrow_pandas.html#map
                    # https://docs.databricks.com/en/pandas/pandas-function-apis.html#map
                    # Source Code: https://github.com/apache/spark/blob/master/python/pyspark/sql/pandas/map_ops.py#L37
                    result_df = json_df.mapInPandas(
                        FhirSenderProcessor.get_process_partition_function(
                            parameters=sender_parameters
                        ),
                        schema=FhirMergeResponseItemSchema.get_schema(),
                    )
                    # execution_plan: str = spark_get_execution_plan(
                    #     df=result_df, extended=True
                    # )

                if result_df is not None:
                    try:
                        if response_path is not None:
                            self.logger.info(
                                f"Executing requests and writing FHIR {resource_name} responses to disk..."
                            )
                            result_df.write.mode(mode).format(file_format).save(
                                str(response_path)
                            )
                            # result_df.show(truncate=False, n=100)
                            self.logger.info(
                                f"Reading from disk and counting rows for {resource_name}..."
                            )
                            result_df = (
                                result_df.sparkSession.read.schema(
                                    FhirMergeResponseItemSchema.get_schema()
                                )
                                .format(file_format)
                                .load(str(response_path))
                            )
                            file_row_count: int = result_df.count()
                            self.logger.info(
                                f"Wrote {file_row_count} FHIR {resource_name} responses to {response_path}"
                            )
                        if view:
                            result_df.createOrReplaceTempView(view)

                        if FhirMergeResponseItemSchema.issue in result_df.columns:
                            # if there are any errors then raise exception
                            first_error_response: Optional[Row] = result_df.where(
                                col(FhirMergeResponseItemSchema.issue).isNotNull()
                            ).first()
                            if first_error_response is not None:
                                if throw_exception_on_validation_failure:
                                    raise FhirSenderValidationException(
                                        url=validation_server_url or server_url,
                                        json_data=json.dumps(
                                            first_error_response[0],
                                            indent=2,
                                            default=str,
                                        ),
                                    )
                                else:
                                    self.logger.info(
                                        f"------- Failed validations for {resource_name} ---------"
                                    )
                                    failed_df = result_df.where(
                                        col(
                                            FhirMergeResponseItemSchema.issue
                                        ).isNotNull()
                                    )
                                    if error_view:
                                        failed_df.createOrReplaceTempView(error_view)
                                    failed_validations: List[str] = (
                                        failed_df.select(
                                            FhirMergeResponseItemSchema.issue
                                        )
                                        .rdd.flatMap(lambda x: x)
                                        .collect()
                                    )
                                    self.logger.info(failed_validations)
                                    self.logger.info(
                                        f"------- End Failed validations for {resource_name} ---------"
                                    )
                            else:
                                if error_view:
                                    df.sparkSession.createDataFrame(
                                        [],
                                        schema=FhirMergeResponseItemSchema.get_schema(),
                                    ).createOrReplaceTempView(error_view)
                        else:
                            if error_view:
                                df.sparkSession.createDataFrame(
                                    [], schema=FhirMergeResponseItemSchema.get_schema()
                                ).createOrReplaceTempView(error_view)
                    except PythonException as e:
                        # noinspection PyUnresolvedReferences
                        if (
                            hasattr(e, "desc")
                            and "pyarrow.lib.ArrowTypeError" in e.desc
                        ):
                            raise FriendlySparkException(
                                exception=e,
                                message="Exception converting data to Arrow format."
                                + f" This is usually because the return data did not match the specified schema.",
                                stage_name=name,
                            )
                        else:
                            raise
                    except Exception as e:
                        raise FriendlySparkException(exception=e, stage_name=name)

        self.logger.info(
            f"----- Finished sending {resource_name} (rows={row_count}) to FHIR server {server_url}  -----"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or f" - {self.getOrDefault(self.resource)}"



import pytest
from typing import List, Dict, Optional

from aioresponses import aioresponses
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor import (
    FhirReceiverProcessor,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.test.fhir_receiver_processor.get_fhir_receiver_parameters import (
    get_fhir_receiver_parameters,
)


@pytest.mark.asyncio
async def test_process_with_token_async() -> None:
    parameters = get_fhir_receiver_parameters()

    resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
        {"resource_id": "1", "access_token": "abc"}
    ]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for result in FhirReceiverProcessor.process_with_token_async(
            partition_index=0,
            resource_id_with_token_list=resource_id_with_token_list,
            parameters=parameters,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']


@pytest.mark.asyncio
async def test_process_one_by_one_async() -> None:
    parameters = get_fhir_receiver_parameters()

    resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
        {"resource_id": "1", "access_token": "abc"}
    ]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for result in FhirReceiverProcessor.process_one_by_one_async(
            partition_index=0,
            first_id="1",
            last_id="1",
            resource_id_with_token_list=resource_id_with_token_list,
            parameters=parameters,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']


@pytest.mark.asyncio
async def test_process_single_row_async() -> None:
    parameters = get_fhir_receiver_parameters()

    resource1: Dict[str, Optional[str]] = {"resource_id": "1", "access_token": "abc"}

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for result in FhirReceiverProcessor.process_single_row_async(
            partition_index=0,
            first_id="1",
            last_id="1",
            resource1=resource1,
            parameters=parameters,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']


@pytest.mark.asyncio
async def test_process_batch_async() -> None:
    parameters = get_fhir_receiver_parameters()

    resource_id_with_token_list: List[Dict[str, Optional[str]]] = [
        {"resource_id": "1", "access_token": "abc"}
    ]

    with aioresponses() as m:
        # Mock the FHIR server response
        m.get(
            "http://fhir-server/Patient/1",
            payload={"resourceType": "Patient", "id": "1"},
        )

        async for result in FhirReceiverProcessor.process_batch_async(
            partition_index=0,
            first_id="1",
            last_id="1",
            resource_id_with_token_list=resource_id_with_token_list,
            parameters=parameters,
        ):
            assert isinstance(result, dict)
            assert result["responses"] == ['{"resourceType": "Patient", "id": "1"}']



from datetime import datetime
from os import environ
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Callable

from helix_fhir_client_sdk.filters.sort_field import SortField
from helix_fhir_client_sdk.function_types import RefreshTokenFunction
from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from pyspark import StorageLevel
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import (
    StructType,
    DataType,
)

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_parameters import (
    FhirReceiverParameters,
)
from spark_pipeline_framework.transformers.fhir_receiver.v2.fhir_receiver_processor_spark import (
    FhirReceiverProcessorSpark,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token_async,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes


class FhirReceiver(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        server_url: Optional[str] = None,
        resource: str,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        id_view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        filter_by_resource: Optional[str] = None,
        filter_parameter: Optional[str] = None,
        additional_parameters: Optional[List[str]] = None,
        additional_parameters_view: Optional[str] = None,
        limit: Optional[int] = None,
        action: Optional[str] = None,
        action_payload: Optional[Dict[str, Any]] = None,
        include_only_properties: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        batch_size: Optional[int] = None,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
        sort_fields: Optional[List[SortField]] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        auth_well_known_url: Optional[str] = None,
        separate_bundle_resources: bool = False,
        error_on_result_count: bool = True,
        expand_fhir_bundle: bool = True,
        accept_type: Optional[str] = None,
        content_type: Optional[str] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        accept_encoding: Optional[str] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        ignore_status_codes: Optional[List[int]] = None,
        verify_counts_match: Optional[bool] = True,
        slug_column: Optional[str] = None,
        url_column: Optional[str] = None,
        error_view: Optional[str] = None,
        view: Optional[str] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        checkpoint_path: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = None,
        use_data_streaming: Optional[bool] = True,
        delta_lake_table: Optional[str] = None,
        schema: Optional[Union[StructType, DataType]] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        graph_json: Optional[Dict[str, Any]] = None,
        run_synchronously: Optional[bool] = None,
        refresh_token_function: Optional[RefreshTokenFunction] = None,
        log_level: Optional[str] = None,
        use_id_above_for_paging: Optional[bool] = True,
        max_chunk_size: int = 100,
        process_chunks_in_parallel: bool = True,
        maximum_concurrent_tasks: int = 100,
        use_uuid_for_id_above: bool = False,
    ) -> None:
        """
        Transformer to call and receive FHIR resources from a FHIR server


        :param server_url: server to call for FHIR
        :param resource: what FHIR resource to retrieve
        :param id_view: view that contains the ids to retrieve (if filter_by_resource is set then this is
                    used in the filter_by_resource query).  If not set, we load all the resources
        :param file_path: where to store the FHIR resources retrieved.  Can be either a path or a function
                    that returns path. Function is passed a loop number if inside a loop or else None
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        :param auth_server_url: server url to call to get the authentication token
        :param filter_by_resource: filter the resource by this. e.g., /Condition?Patient=1
                (resource=Condition, filter_by_resource=Patient)
        :param filter_parameter: Instead of requesting ?patient=1,
                do ?subject:Patient=1 (if filter_parameter is subject)
        :param additional_parameters: Any additional parameters to send with request
        :param additional_parameters_view: (Optional) view name for additional parameters, e.g. search parameters
        :param auth_scopes: list of scopes to request permission for e.g., system/AllergyIntolerance.read
        :param limit: maximum number of resources to get
        :param action: (Optional) do an action e.g., $everything
        :param action_payload: (Optional) in case action needs a http request payload
        :param page_size: (Optional) use paging and get this many items in each page
        :param batch_size: (Optional) How many id rows to send to FHIR server in one call
        :param last_updated_after: (Optional) Only get records newer than this
        :param last_updated_before: (Optional) Only get records older than this
        :param sort_fields: sort by fields in the resource
        :param separate_bundle_resources: True means we group the response by FHIR resources
                                            (check helix.fhir.client.sdk/tests/test_fhir_client_bundle.py)
        :param error_on_result_count: Throw exception when the response from FHIR count does not match the request count
        :param accept_type: (Optional) Accept header to use
        :param content_type: (Optional) Content-Type header to use
        :param additional_request_headers: (Optional) Additional request headers to use
                                            (Eg: {"Accept-Charset": "utf-8"})
        :param accept_encoding: (Optional) Accept-encoding header to use
        :param ignore_status_codes: (Optional) do not throw an exception for these HTTP status codes
        :param mode: if output files exist, should we overwrite or append
        :param slug_column: (Optional) use this column to set the security tags
        :param url_column: (Optional) column to read the url
        :param error_view: (Optional) log errors into this view (view only exists IF there are errors) and
                            don't throw exceptions.  schema: url, error_text, status_code
        :param use_data_streaming: (Optional) whether to use data streaming i.e., HTTP chunk transfer encoding
        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param schema: the schema to apply after we receive the data
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param graph_json: (Optional) a FHIR GraphDefinition resource to use for retrieving data
        :param run_synchronously: (Optional) Run on the Spark master to make debugging easier on dev machines
        :param refresh_token_function: (Optional) function to refresh the token
        :param use_uuid_for_id_above: (boolean) flag indicating whether to use UUIDs for the "id:above" parameter when
         fetching next resources.
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)

        assert not limit or isinstance(
            limit, int
        ), f"limit is not an int. it is a {type(limit)}"
        assert not batch_size or isinstance(
            batch_size, int
        ), f"batch_size is not an int. it is a {type(batch_size)}"
        assert not page_size or isinstance(
            page_size, int
        ), f"page_size is not an int. it is a {type(page_size)}"

        assert not last_updated_after or isinstance(
            last_updated_after, datetime
        ), f"last_updated_after is not a datetime. it is a {type(last_updated_after)}"

        assert not last_updated_before or isinstance(
            last_updated_before, datetime
        ), f"last_updated_before is not a datetime. it is a {type(last_updated_before)}"

        assert file_path

        if action == "$graph":
            assert id_view, "id_view is required when action is $graph"
            assert action_payload, "action_payload is required when action is $graph"
            assert not view, (
                "view cannot be specified if action is $graph since $graph returns"
                " many different resource types."
                "  Use FhirReader to read the file_path and specify the schema."
            )
            assert (
                additional_parameters
            ), "additional_parameters must be specified if action is $graph"
            assert [
                a for a in additional_parameters if a.startswith("contained")
            ], "additional_parameters must contain 'contained' when action is $graph"

        self.log_level: Param[str] = Param(self, "log_level", "")
        self._setDefault(log_level=log_level)

        self.logger = get_logger(__name__, level=log_level or "INFO")

        self.server_url: Param[Optional[str]] = Param(self, "server_url", "")
        self._setDefault(server_url=None)

        self.resource: Param[str] = Param(self, "resource", "")
        self._setDefault(resource=None)

        self.filter_by_resource: Param[Optional[str]] = Param(
            self, "filter_by_resource", ""
        )
        self._setDefault(filter_by_resource=None)

        self.filter_parameter: Param[Optional[str]] = Param(
            self, "filter_parameter", ""
        )
        self._setDefault(filter_parameter=None)

        self.additional_parameters: Param[Optional[List[str]]] = Param(
            self, "additional_parameters", ""
        )
        self._setDefault(additional_parameters=None)

        self.additional_parameters_view: Param[Optional[str]] = Param(
            self, "additional_parameters_view", ""
        )
        self._setDefault(additional_parameters_view=None)

        self.id_view: Param[Optional[str]] = Param(self, "id_view", "")
        self._setDefault(id_view=None)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.action: Param[Optional[str]] = Param(self, "action", "")
        self._setDefault(action=None)

        self.action_payload: Param[Optional[Dict[str, Any]]] = Param(
            self, "action_payload", ""
        )
        self._setDefault(action_payload=None)

        self.include_only_properties: Param[Optional[List[str]]] = Param(
            self, "include_only_properties", ""
        )
        self._setDefault(include_only_properties=None)

        self.page_size: Param[Optional[int]] = Param(self, "page_size", "")
        self._setDefault(page_size=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=None)

        self.last_updated_after: Param[Optional[datetime]] = Param(
            self, "last_updated_after", ""
        )
        self._setDefault(last_updated_after=None)

        self.last_updated_before: Param[Optional[datetime]] = Param(
            self, "last_updated_before", ""
        )
        self._setDefault(last_updated_before=None)

        self.sort_fields: Param[Optional[List[SortField]]] = Param(
            self, "sort_fields", ""
        )
        self._setDefault(sort_fields=None)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[Optional[List[str]]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)

        self.auth_well_known_url: Param[Optional[str]] = Param(
            self, "auth_well_known_url", ""
        )
        self._setDefault(auth_well_known_url=None)

        self.separate_bundle_resources: Param[bool] = Param(
            self, "separate_bundle_resources", ""
        )
        self._setDefault(separate_bundle_resources=False)

        self.expand_fhir_bundle: Param[bool] = Param(self, "expand_fhir_bundle", "")
        self._setDefault(expand_fhir_bundle=expand_fhir_bundle)

        self.error_on_result_count: Param[bool] = Param(
            self, "error_on_result_count", ""
        )
        self._setDefault(error_on_result_count=error_on_result_count)

        self.accept_type: Param[Optional[str]] = Param(self, "accept_type", "")
        self._setDefault(accept_type=accept_type)

        self.content_type: Param[Optional[str]] = Param(self, "content_type", "")
        self._setDefault(content_type=content_type)

        self.additional_request_headers: Param[Optional[Dict[str, str]]] = Param(
            self, "additional_request_headers", ""
        )
        self._setDefault(additional_request_headers=additional_request_headers)

        self.accept_encoding: Param[Optional[str]] = Param(self, "accept_encoding", "")
        self._setDefault(accept_encoding=accept_encoding)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.ignore_status_codes: Param[Optional[List[int]]] = Param(
            self, "ignore_status_codes", ""
        )
        self._setDefault(ignore_status_codes=None)

        self.verify_counts_match: Param[Optional[bool]] = Param(
            self, "verify_counts_match", ""
        )
        self._setDefault(verify_counts_match=None)

        self.slug_column: Param[Optional[str]] = Param(self, "slug_column", "")
        self._setDefault(slug_column=None)

        self.url_column: Param[Optional[str]] = Param(self, "url_column", "")
        self._setDefault(url_column=None)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.retry_count: Param[Optional[int]] = Param(self, "retry_count", "")
        self._setDefault(retry_count=None)

        self.exclude_status_codes_from_retry: Param[Optional[List[int]]] = Param(
            self, "exclude_status_codes_from_retry", ""
        )
        self._setDefault(exclude_status_codes_from_retry=None)

        self.checkpoint_path: Param[
            Optional[Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]]
        ] = Param(self, "checkpoint_path", "")
        self._setDefault(checkpoint_path=None)

        self.use_data_streaming: Param[Optional[bool]] = Param(
            self, "use_data_streaming", ""
        )
        self._setDefault(use_data_streaming=True)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=None)

        if delta_lake_table:
            assert schema is not None, f"schema must be supplied when using delta talk"

        self.schema: Param[Optional[Union[StructType, DataType]]] = Param(
            self, "schema", ""
        )
        self._setDefault(schema=None)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.graph_json: Param[Optional[Dict[str, Any]]] = Param(self, "graph_json", "")
        self._setDefault(graph_json=None)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.refresh_token_function: Param[Optional[RefreshTokenFunction]] = Param(
            self, "refresh_token_function", ""
        )
        self._setDefault(refresh_token_function=refresh_token_function)

        self.use_id_above_for_paging: Param[Optional[bool]] = Param(
            self, "use_id_above_for_paging", ""
        )
        self._setDefault(use_id_above_for_paging=use_id_above_for_paging)

        self.max_chunk_size: Param[int] = Param(self, "max_chunk_size", "")
        self._setDefault(max_chunk_size=max_chunk_size)

        self.process_chunks_in_parallel: Param[bool] = Param(
            self, "process_chunks_in_parallel", ""
        )
        self._setDefault(process_chunks_in_parallel=process_chunks_in_parallel)

        self.maximum_concurrent_tasks: Param[int] = Param(
            self, "maximum_concurrent_tasks", ""
        )
        self._setDefault(maximum_concurrent_tasks=maximum_concurrent_tasks)

        self.use_uuid_for_id_above: Param[bool] = Param(
            self, "use_uuid_for_id_above", ""
        )
        self._setDefault(use_uuid_for_id_above=use_uuid_for_id_above)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    async def _transform_async(self, df: DataFrame) -> DataFrame:
        server_url: Optional[str] = self.getServerUrl()
        resource_name: str = self.getResource()
        parameters = self.getParameters()
        filter_by_resource: Optional[str] = self.getFilterByResource()
        filter_parameter: Optional[str] = self.getFilterParameter()
        additional_parameters: Optional[List[str]] = self.getAdditionalParameters()
        additional_parameters_view: Optional[str] = self.getOrDefault(
            self.additional_parameters_view
        )
        id_view: Optional[str] = self.getIdView()
        file_path1: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        file_path: Union[Path, str] = (
            file_path1(self.loop_id) if callable(file_path1) else file_path1
        )
        name: Optional[str] = self.getName()
        action: Optional[str] = self.getAction()
        action_payload: Optional[Dict[str, Any]] = self.getActionPayload()
        include_only_properties: Optional[List[str]] = self.getIncludeOnlyProperties()
        page_size: Optional[int] = self.getPageSize()
        batch_size: Optional[int] = self.getBatchSize()
        limit: Optional[int] = self.getLimit()
        last_updated_before: Optional[datetime] = self.getLastUpdatedBefore()
        last_updated_after: Optional[datetime] = self.getLastUpdateAfter()
        sort_fields: Optional[List[SortField]] = self.getSortFields()

        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()
        auth_well_known_url: Optional[str] = self.getOrDefault(self.auth_well_known_url)

        separate_bundle_resources: bool = self.getSeparateBundleResources()
        expand_fhir_bundle: bool = self.getExpandFhirBundle()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        auth_access_token: Optional[str] = None

        error_on_result_count: bool = self.getErrorOnResultCount()

        accept_type: Optional[str] = self.getAcceptType()
        content_type: Optional[str] = self.getContentType()
        accept_encoding: Optional[str] = self.getAcceptEncoding()
        additional_request_headers: Optional[Dict[str, str]] = (
            self.getAdditionalRequestHeaders()
        )

        ignore_status_codes: List[int] = self.getIgnoreStatusCodes() or []
        ignore_status_codes.append(200)

        verify_counts_match: Optional[bool] = self.getVerifyCountsMatch()
        mode: str = self.getMode()

        slug_column: Optional[str] = self.getOrDefault(self.slug_column)
        url_column: Optional[str] = self.getOrDefault(self.url_column)

        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view: Optional[str] = self.getOrDefault(self.view)

        retry_count: Optional[int] = self.getOrDefault(self.retry_count)
        exclude_status_codes_from_retry: Optional[List[int]] = self.getOrDefault(
            self.exclude_status_codes_from_retry
        )

        checkpoint_path1: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = self.getOrDefault(self.checkpoint_path)
        checkpoint_path: Path | str | None = (
            checkpoint_path1(self.loop_id)
            if checkpoint_path1 and callable(checkpoint_path1)
            else checkpoint_path1
        )

        log_level: Optional[str] = self.getOrDefault(self.log_level) or environ.get(
            "LOGLEVEL"
        )

        use_data_streaming: Optional[bool] = self.getOrDefault(self.use_data_streaming)

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        schema: Optional[Union[StructType, DataType]] = self.getOrDefault(self.schema)

        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )

        graph_json: Optional[Dict[str, Any]] = self.getOrDefault(self.graph_json)

        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)

        refresh_token_function: Optional[RefreshTokenFunction] = self.getOrDefault(
            self.refresh_token_function
        )
        use_id_above_for_paging: Optional[bool] = self.getOrDefault(
            self.use_id_above_for_paging
        )

        max_chunk_size: int = self.getOrDefault(self.max_chunk_size)
        process_chunks_in_parallel: bool = self.getOrDefault(
            self.process_chunks_in_parallel
        )
        maximum_concurrent_tasks: int = self.getOrDefault(self.maximum_concurrent_tasks)
        use_uuid_for_id_above: bool = self.getOrDefault(self.use_uuid_for_id_above)

        if parameters and parameters.get("flow_name"):
            if additional_request_headers:
                additional_request_headers.update(
                    {
                        "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                    }
                )
            else:
                additional_request_headers = {
                    "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                }

        # get access token first so we can reuse it
        if auth_client_id and server_url:
            auth_access_token_result: GetAccessTokenResult = (
                await fhir_get_access_token_async(
                    logger=self.logger,
                    server_url=server_url,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    log_level=log_level,
                    auth_well_known_url=auth_well_known_url,
                )
            )
            auth_access_token = auth_access_token_result.access_token

        with ProgressLogMetric(
            name=f"{name}_fhir_receiver", progress_logger=progress_logger
        ):
            # if additional_parameters_view is given,
            # this is for generic use. e.g. to support blocking by adding search parameters in a query to FHIR
            if additional_parameters_view and additional_parameters:
                # created a df for additional_parameters_view
                # the data in this view MUST have 1 row per each parameter and only 1 column per each row.
                # e.g. "address-postalcode=10304,11040,11801" for a comma separate postal codes for Person resource
                df_additional_parameters_view: DataFrame = df.sparkSession.table(
                    additional_parameters_view
                )
                for row in df_additional_parameters_view.collect():
                    # get the data from each row for each parameter, e.g. "address-postalcode=10304,11040,11801"
                    parameter_values: str = str(row[0])
                    # add it to the additional_parameters
                    additional_parameters += [f"{parameter_values}"]

            receiver_parameters: FhirReceiverParameters = FhirReceiverParameters(
                total_partitions=None,  # will be calculated later
                batch_size=batch_size,
                has_token_col=False,  # will be calculated later
                server_url=server_url,
                log_level=log_level,
                action=action,
                action_payload=action_payload,
                additional_parameters=additional_parameters,
                filter_by_resource=filter_by_resource,
                filter_parameter=filter_parameter,
                sort_fields=sort_fields,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                auth_well_known_url=auth_well_known_url,
                include_only_properties=include_only_properties,
                separate_bundle_resources=separate_bundle_resources,
                expand_fhir_bundle=expand_fhir_bundle,
                accept_type=accept_type,
                content_type=content_type,
                additional_request_headers=additional_request_headers,
                accept_encoding=accept_encoding,
                slug_column=slug_column,
                retry_count=retry_count,
                exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                limit=limit,
                auth_access_token=auth_access_token,
                resource_type=resource_name,
                error_view=error_view,
                url_column=url_column,
                use_data_streaming=use_data_streaming,
                graph_json=graph_json,
                ignore_status_codes=ignore_status_codes,
                refresh_token_function=refresh_token_function,
                use_id_above_for_paging=use_id_above_for_paging,
                pandas_udf_parameters=AsyncPandasUdfParameters(
                    max_chunk_size=max_chunk_size,
                    process_chunks_in_parallel=process_chunks_in_parallel,
                    log_level=log_level,
                    maximum_concurrent_tasks=maximum_concurrent_tasks,
                ),
                use_uuid_for_id_above=use_uuid_for_id_above,
            )

            if id_view:
                return await FhirReceiverProcessorSpark.get_resources_by_id_view_async(
                    df=df,
                    id_view=id_view,
                    parameters=receiver_parameters,
                    view=view,
                    run_synchronously=run_synchronously,
                    checkpoint_path=checkpoint_path,
                    progress_logger=progress_logger,
                    delta_lake_table=delta_lake_table,
                    cache_storage_level=cache_storage_level,
                    error_view=error_view,
                    name=name,
                    expand_fhir_bundle=expand_fhir_bundle,
                    file_path=file_path,
                    schema=schema,
                    mode=mode,
                    error_on_result_count=error_on_result_count,
                    verify_counts_match=verify_counts_match,
                    logger=self.logger,
                )
            else:  # get all resources
                return await FhirReceiverProcessorSpark.get_all_resources_async(
                    df=df,
                    parameters=receiver_parameters,
                    delta_lake_table=delta_lake_table,
                    last_updated_after=last_updated_after,
                    last_updated_before=last_updated_before,
                    batch_size=batch_size,
                    mode=mode,
                    file_path=file_path,
                    page_size=page_size,
                    limit=limit,
                    progress_logger=progress_logger,
                    view=view,
                    error_view=error_view,
                    logger=self.logger,
                    schema=schema,
                )

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResource(self) -> str:
        return self.getOrDefault(self.resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilterByResource(self) -> Optional[str]:
        return self.getOrDefault(self.filter_by_resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilterParameter(self) -> Optional[str]:
        return self.getOrDefault(self.filter_parameter)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalParameters(self) -> Optional[List[str]]:
        return self.getOrDefault(self.additional_parameters)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIdView(self) -> Optional[str]:
        return self.getOrDefault(self.id_view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAction(self) -> Optional[str]:
        return self.getOrDefault(self.action)

    # noinspection PyPep8Naming
    def getActionPayload(self) -> Optional[Dict[str, Any]]:
        return self.getOrDefault(self.action_payload)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIncludeOnlyProperties(self) -> Optional[List[str]]:
        return self.getOrDefault(self.include_only_properties)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getPageSize(self) -> Optional[int]:
        return self.getOrDefault(self.page_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> Optional[int]:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLastUpdateAfter(self) -> Optional[datetime]:
        return self.getOrDefault(self.last_updated_after)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLastUpdatedBefore(self) -> Optional[datetime]:
        return self.getOrDefault(self.last_updated_before)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSortFields(self) -> Optional[List[SortField]]:
        return self.getOrDefault(self.sort_fields)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSeparateBundleResources(self) -> bool:
        return self.getOrDefault(self.separate_bundle_resources)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getExpandFhirBundle(self) -> bool:
        return self.getOrDefault(self.expand_fhir_bundle)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getErrorOnResultCount(self) -> bool:
        return self.getOrDefault(self.error_on_result_count)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAcceptType(self) -> Optional[str]:
        return self.getOrDefault(self.accept_type)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getContentType(self) -> Optional[str]:
        return self.getOrDefault(self.content_type)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalRequestHeaders(self) -> Optional[Dict[str, str]]:
        return self.getOrDefault(self.additional_request_headers)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAcceptEncoding(self) -> Optional[str]:
        return self.getOrDefault(self.accept_encoding)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIgnoreStatusCodes(self) -> Optional[List[int]]:
        return self.getOrDefault(self.ignore_status_codes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getVerifyCountsMatch(self) -> Optional[bool]:
        return self.getOrDefault(self.verify_counts_match)



from typing import Dict, Any, Optional, List

from pyspark.ml.param import Param
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)

# noinspection PyPep8Naming
from pyspark.sql import functions as F
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_table_exists,
)


class FrameworkGroupByTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        group_by_columns: List[str],
        source_view: str,
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        append_to_view: Optional[bool] = None,
        concatenate_content_as_json: Optional[bool] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert source_view
        assert view
        assert group_by_columns is not None
        assert len(group_by_columns) > 0

        self.logger = get_logger(__name__)

        self.group_by_column: Param[str] = Param(self, "group_by_column", "")
        self._setDefault(group_by_column=group_by_columns)

        self.source_view: Param[str] = Param(self, "source_view", "")
        self._setDefault(source_view=source_view)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.append_to_view: Param[Optional[bool]] = Param(self, "append_to_view", "")
        self._setDefault(append_to_view=append_to_view)

        self.concatenate_content_as_json: Param[Optional[bool]] = Param(
            self, "concatenate_content_as_json", ""
        )
        self._setDefault(concatenate_content_as_json=concatenate_content_as_json)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        source_view: str = self.getOrDefault(self.source_view)
        view: str = self.getOrDefault(self.view)
        append_to_view: Optional[bool] = self.getOrDefault(self.append_to_view)
        concatenate_content_as_json: Optional[bool] = self.getOrDefault(
            self.concatenate_content_as_json
        )
        group_by_column: str = self.getOrDefault(self.group_by_column)

        df_grouped: DataFrame
        if concatenate_content_as_json:
            df_grouped = df.groupBy(*group_by_column).agg(
                F.count("*").alias("count"),
                F.concat_ws(
                    ",",
                    F.collect_list(F.to_json(F.struct([df[x] for x in df.columns]))),
                ).alias("json"),
            )
        else:
            df_grouped = df.groupBy(group_by_column).agg(F.count("*").alias("count"))

        if append_to_view:
            if spark_table_exists(df.sparkSession, view):
                df_grouped = df.sparkSession.table(view).union(df_grouped)

        df_grouped.createOrReplaceTempView(view)

        return df



from typing import (
    Optional,
    Any,
    Callable,
    Tuple,
    Awaitable,
    AsyncGenerator,
)

from aiohttp import ClientResponse
from pyspark.sql import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.api_helper.v2.http_request import (
    HelixHttpRequest,
)


RESPONSE_PROCESSOR_TYPE = Callable[
    [ClientResponse, Any],
    Awaitable[Tuple[Any, bool]],
]

REQUEST_GENERATOR_TYPE = Callable[
    [DataFrame, Optional[ProgressLogger]],
    AsyncGenerator[Tuple[HelixHttpRequest, Any], None],
]



from os import path, makedirs, environ
from pathlib import Path
from shutil import rmtree

from mockserver_client.mock_requests_loader import (
    load_mock_fhir_requests_from_folder,
    load_mock_source_api_json_responses,
)
from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_receiver.v1.fhir_receiver import (
    FhirReceiver,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)
from mockserver_client.mockserver_client import MockServerFriendlyClient

slot_practitioner_graph = {
    "resourceType": "GraphDefinition",
    "id": "o",
    "name": "provider_slots",
    "status": "active",
    "start": "Practitioner",
    "link": [
        {
            "target": [
                {
                    "type": "PractitionerRole",
                    "params": "practitioner={ref}",
                    "link": [
                        {
                            "target": [
                                {
                                    "type": "Schedule",
                                    "params": "actor={ref}",
                                    "link": [
                                        {
                                            "target": [
                                                {
                                                    "type": "Slot",
                                                    "params": "schedule={ref}",
                                                }
                                            ]
                                        }
                                    ],
                                }
                            ]
                        }
                    ],
                }
            ]
        }
    ],
}


def test_fhir_receiver_graph_synthetic_run_synchronously(
    spark_session: SparkSession,
) -> None:
    # Arrange
    print()

    environ["LOGLEVEL"] = "DEBUG"
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    patient_json_path: Path = temp_folder.joinpath("patient_graph.json")

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    spark_session.createDataFrame(
        [
            ("01-practitioner", "Qureshi"),
            ("02-practitioner", "Vidal"),
        ],
        ["id", "name"],
    ).createOrReplaceTempView("fhir_ids")

    test_name = "test_fhir_receiver_graph"
    fhir_server_url = f"http://mock-server:1080/{test_name}/4_0_0"
    client = MockServerFriendlyClient("http://mock-server:1080")
    client.clear(test_name)

    load_mock_fhir_requests_from_folder(
        folder=data_dir.joinpath("fhir_graph_calls_synthetic/practitioner"),
        mock_client=client,
        method="GET",
        url_prefix=f"{test_name}",
    )

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("fhir_graph_calls_synthetic/practitionerRole"),
        mock_client=client,
        url_prefix=f"{test_name}/4_0_0/PractitionerRole",
    )

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("fhir_graph_calls_synthetic/schedule"),
        mock_client=client,
        url_prefix=f"{test_name}/4_0_0/Schedule",
    )

    # Act
    with ProgressLogger() as progress_logger:
        FhirReceiver(
            server_url=fhir_server_url,
            resource="Practitioner",
            id_view="fhir_ids",
            file_path=patient_json_path,
            progress_logger=progress_logger,
            # action="$graph",
            page_size=2,
            limit=2,
            batch_size=2,
            # additional_parameters=["contained=true"],
            # expand_fhir_bundle=True,
            graph_json=slot_practitioner_graph,
            separate_bundle_resources=True,
            run_synchronously=True,
        ).transform(df)

    # Assert
    json_df: DataFrame = df.sparkSession.read.json(str(patient_json_path))
    json_df.show(truncate=False)
    json_df.printSchema()

    practitioners = json_df.select("Practitioner").collect()[0][0]
    assert len(practitioners) == 1
    assert practitioners[0].asDict() == {
        "id": "01-practitioner",
        "resourceType": "Practitioner",
    }

    practitioner_roles = json_df.select("PractitionerRole").collect()[0][0]
    assert len(practitioner_roles) == 1
    assert practitioner_roles[0]["id"] == "4657-3437"

    schedules = json_df.select("Schedule").collect()[0][0]
    assert len(schedules) == 1
    assert schedules[0]["id"] == "1720233406-SCH-MPCS"



import json
from typing import Optional

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)


class OAuth2ClientCredentialsFlow:
    def __init__(
        self,
        *,
        auth_url: str,
        client_id: str,
        client_secret: str,
        progress_logger: Optional[ProgressLogger],
    ) -> None:
        """
        Encapsulate the logic to connect to OAuth2 servers using client credentials flow



        """
        self.auth_url: str = auth_url
        assert auth_url
        self.client_id: str = client_id
        assert client_id
        self.client_secret: str = client_secret
        assert client_secret
        self.progress_logger: Optional[ProgressLogger] = progress_logger

    def get_token(self) -> Optional[str]:
        http_request = HelixHttpRequest(
            url=self.auth_url,
            payload={
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "grant_type": "client_credentials",
            },
            request_type=RequestType.POST,
        )

        response = http_request.get_result()
        if self.progress_logger:
            self.progress_logger.write_to_log(
                f"Received from {self.auth_url}: {json.dumps(response.result)}"
            )
        token = response.result.get("access_token")

        return token



import logging

from aioresponses import aioresponses
from typing import List, Dict, Any
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_helpers import (
    ElasticSearchHelpers,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v2.elasticsearch_result import (
    ElasticSearchResult,
)


async def test_send_json_bundle_to_elasticsearch_async() -> None:
    json_data_list: List[str] = [
        '{"id": "1", "name": "test1"}',
        '{"id": "2", "name": "test2"}',
    ]
    index = "test_index"
    operation = "index"
    doc_id_prefix = "prefix"
    timeout = 60

    with aioresponses() as m:
        # Mock the async_bulk call
        return_payload = {
            "items": [
                {
                    "index": {
                        "_index": "test",
                        "_id": "1",
                        "_version": 1,
                        "result": "created",
                        "_shards": {"total": 2, "successful": 2, "failed": 0},
                        "status": 201,
                        "_seq_no": 0,
                        "_primary_term": 1,
                    }
                },
                {
                    "index": {
                        "_index": "test",
                        "_id": "2",
                        "_version": 1,
                        "result": "created",
                        "_shards": {"total": 2, "successful": 2, "failed": 0},
                        "status": 201,
                        "_seq_no": 0,
                        "_primary_term": 1,
                    }
                },
            ],
            "errors": False,
        }
        m.post(
            "https://elasticsearch:9200/test_index/_bulk",
            status=200,
            payload=return_payload,
        )

        result: ElasticSearchResult = (
            await ElasticSearchHelpers.send_json_bundle_to_elasticsearch_async(
                json_data_list=json_data_list,
                index=index,
                operation=operation,
                logger=logging.getLogger(__name__),
                doc_id_prefix=doc_id_prefix,
                timeout=timeout,
            )
        )

        assert result.success == 2, f"result: {result}"
        assert result.failed == 0, f"result: {result}"
        assert (
            result.url == "https://elasticsearch:9200/test_index"
        ), f"result: {result}"
        assert result.error is None, f"result: {result}"


async def test_send_json_bundle_to_real_elasticsearch_async() -> None:
    json_data_list: List[str] = [
        '{"id": "1", "name": "test1"}',
        '{"id": "2", "name": "test2"}',
    ]
    index = "test_index"
    operation = "index"
    doc_id_prefix = "prefix"
    timeout = 60

    result: ElasticSearchResult = (
        await ElasticSearchHelpers.send_json_bundle_to_elasticsearch_async(
            json_data_list=json_data_list,
            index=index,
            operation=operation,
            logger=logging.getLogger(__name__),
            doc_id_prefix=doc_id_prefix,
            timeout=timeout,
        )
    )

    assert result.success == 2, f"result: {result}"
    assert result.failed == 0, f"result: {result}"
    assert result.url == "https://elasticsearch:9200/test_index", f"result: {result}"
    assert result.error is None, f"result: {result}"


def test_generate_es_bulk() -> None:
    print()
    json_data_list: List[str] = [
        '{"id": "1", "name": "test1"}',
        '{"id": "2", "name": "test2"}',
    ]
    index = "test_index"
    operation = "index"
    doc_id_prefix = "prefix"

    bulk_data: List[Dict[str, Any]] = list(
        ElasticSearchHelpers.generate_es_bulk(
            json_data_list=json_data_list,
            operation=operation,
            index=index,
            doc_id_prefix=doc_id_prefix,
        )
    )

    print(f"bulk_data: {bulk_data}")
    assert len(bulk_data) == 2
    assert bulk_data[0] == {"id": "1", "name": "test1", "_id": "prefix-1"}
    assert bulk_data[1] == {"id": "2", "name": "test2", "_id": "prefix-2"}



import json
import os
import random
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, cast

import aiohttp
import boto3
import pymongo
import pytest
from aioresponses import aioresponses
from moto import mock_aws
from pymongo.collection import Collection

from spark_pipeline_framework.utilities.helix_geolocation.v2.address_standardizer import (
    AddressStandardizer,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.document_db_cache_handler import (
    DocumentDBCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.mock_cache_handler import (
    MockCacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendor_response_key_error import (
    VendorResponseKeyError,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.melissa_standardizing_vendor import (
    MelissaStandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.mock_standardizing_vendor import (
    MockStandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)

mongo_address = r"mongodb://mongo:27017/"
mongo_collection_name = "helix_address_cache"

raw_addr_obj = RawAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    line2=None,
    city="Austin",
    state="TX",
    zipcode="78753",
)

raw_addr_obj1 = RawAddress(
    address_id="100",
    line1="8300 N Lamar Blvd",
    line2=None,
    city="Austin",
    state="TX",
    zipcode="78753",
)

raw_addr_obj2 = RawAddress(
    address_id="11",
    line1="1137 Huntington Drive Ste B1, South Pasadena, CA 91030",
    country="US",
    line2=None,
    city=None,
    state=None,
    zipcode=None,
)

std_addr_obj = StandardizedAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    county=None,
    country="US",
    zipcode="78753",
    latitude="30.373400",
    longitude="-97.680000",
    formatted_address="8300 North Lamar Boulevard;Austin, TX 78753-5976",
    standardize_vendor="melissa",
)

std_addr_obj1 = StandardizedAddress(
    address_id="10",
    line1="8300 N Lamar Blvd",
    city="Austin",
    state="TX",
    county=None,
    country="US",
    zipcode="78753",
    latitude="30.373400",
    longitude="-97.680000",
    formatted_address="8300 North Lamar Boulevard;Austin, TX 78753-5976",
    standardize_vendor="melissa",
)

std_addr_obj2 = StandardizedAddress(
    address_id="11",
    line1="1137 Huntington Drive Ste B1",
    city="Los Angeles",
    state="CA",
    county=None,
    country="US",
    zipcode="91030-4582",
    latitude="34.100066",
    longitude="-118.155034",
    formatted_address="1137 Huntington Dr Ste B1;South Pasadena, CA 91030-4582",
    standardize_vendor="melissa",
)
response_data: Dict[str, Any] = {
    "Version": "3.0.1.160",
    "TransmissionReference": "mock test",
    "TransmissionResults": "",
    "TotalRecords": "2",
    "Records": [
        {
            "RecordID": "11",
            "Results": "AC11,AC16,AV25,GS06",
            "FormattedAddress": "1137 Huntington Dr Ste B1;South Pasadena, CA 91030-4582",
            "Organization": "",
            "AddressLine1": "1137 Huntington Dr Ste B1",
            "AddressLine2": "South Pasadena, CA 91030-4582",
            "AddressLine3": "",
            "AddressLine4": "",
            "AddressLine5": "",
            "AddressLine6": "",
            "AddressLine7": "",
            "SubPremises": "Ste B1",
            "DoubleDependentLocality": "",
            "DependentLocality": "",
            "Locality": "South Pasadena",
            "SubAdministrativeArea": "Los Angeles",
            "AdministrativeArea": "CA",
            "PostalCode": "91030-4582",
            "PostalCodeType": "",
            "AddressType": "H",
            "AddressKey": "91030458221",
            "SubNationalArea": "",
            "CountryName": "United States of America",
            "CountryISO3166_1_Alpha2": "US",
            "CountryISO3166_1_Alpha3": "USA",
            "CountryISO3166_1_Numeric": "840",
            "CountrySubdivisionCode": "US-CA",
            "Thoroughfare": "Huntington Drive",
            "ThoroughfarePreDirection": "",
            "ThoroughfareLeadingType": "",
            "ThoroughfareName": "Huntington",
            "ThoroughfareTrailingType": "Drive",
            "ThoroughfarePostDirection": "",
            "DependentThoroughfare": "",
            "DependentThoroughfarePreDirection": "",
            "DependentThoroughfareLeadingType": "",
            "DependentThoroughfareName": "",
            "DependentThoroughfareTrailingType": "",
            "DependentThoroughfarePostDirection": "",
            "Building": "",
            "PremisesType": "",
            "PremisesNumber": "1137",
            "SubPremisesType": "Ste",
            "SubPremisesNumber": "B1",
            "PostBox": "",
            "Latitude": "34.100066",
            "Longitude": "-118.155034",
            "DeliveryIndicator": "B",
            "MelissaAddressKey": "7245851331",
            "MelissaAddressKeyBase": "7381043882",
            "PostOfficeLocation": "",
            "SubPremiseLevel": "",
            "SubPremiseLevelType": "",
            "SubPremiseLevelNumber": "",
            "SubBuilding": "",
            "SubBuildingType": "",
            "SubBuildingNumber": "",
            "UTC": "UTC-08:00",
            "DST": "Y",
            "DeliveryPointSuffix": "",
            "CensusKey": "060374807042021",
            "Extras": "",
        }
    ],
}


def test_raw_address_to_export() -> None:
    # addr_dict = raw_addr_obj.to_dict()
    addr_str = raw_addr_obj.to_str()
    addr_hash = raw_addr_obj.to_hash()

    # commenting out because the internal_id changes each time
    # assert addr_dict == {
    #     "address_id": "10",
    #     "country": "US",
    #     "zipcode": "78753",
    #     "state": "TX",
    #     "city": "Austin",
    #     "line2": None,
    #     "line1": "8300 N Lamar Blvd",
    # }
    assert addr_str == "8300 N Lamar Blvd, Austin TX 78753 US"
    assert addr_hash == "aa866e8b20823e573dba7c369b778196d7e75208"


def test_std_address_to_export() -> None:
    # addr_dict = std_addr_obj.to_dict()
    addr_str = std_addr_obj.formatted_address

    # commenting out because the internal_id changes each time
    # assert addr_dict == {
    #     "address_id": "10",
    #     "country": "US",
    #     "zipcode": "78753",
    #     "state": "TX",
    #     "city": "Austin",
    #     "line2": None,
    #     "line1": "8300 N Lamar Blvd",
    #     "county": None,
    #     "latitude": "30.373400",
    #     "longitude": "-97.680000",
    #     "formatted_address": "8300 North Lamar Boulevard;Austin, TX 78753-5976",
    #     "standardize_vendor": "melissa",
    # }
    assert addr_str == "8300 North Lamar Boulevard;Austin, TX 78753-5976"


def read_cache_file() -> List[Dict[str, Any]]:
    test_file_path = Path(__file__).parent.joinpath("cache_test_data.json")
    with open(test_file_path) as f:
        r: List[Dict[str, Any]] = json.load(f)
        return r


async def test_address_api_call() -> None:
    with mock_aws():
        with aioresponses() as m:
            # arrange
            os.environ["AWS_ACCESS_KEY_ID"] = "testing"
            os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
            os.environ["AWS_SECURITY_TOKEN"] = "testing"
            os.environ["AWS_SESSION_TOKEN"] = "testing"
            ssm = boto3.client("ssm", region_name="us-east-1")
            ssm.put_parameter(
                Name="/prod/helix/external/melissa/license_key_credit",
                Value="fake license",
                Type="SecureString",
            )
            raw_addr_obj_copy = deepcopy(raw_addr_obj)
            raw_addr_obj_copy.set_id("100")
            raw_addrs = [raw_addr_obj_copy, raw_addr_obj1, raw_addr_obj2]
            expected_result = [std_addr_obj, std_addr_obj1, std_addr_obj2]
            arrange_mongo()
            m.post("https://api.melissa.com/v3/standardize", payload=response_data)
            # act
            r: List[
                StandardizedAddress
            ] = await AddressStandardizer().standardize_async(
                raw_addrs,
                vendor_obj=cast(
                    StandardizingVendor[BaseVendorApiResponse],
                    MockStandardizingVendor(),
                ),
                cache_handler_obj=DocumentDBCacheHandler(),
            )
            assert 3 == len(r)
            # assert
            assert r[0].get_id() == "100"
            assert r[1].get_id() == "100"
            assert r[2].get_id() == "11"
            assert r[0].formatted_address == expected_result[0].formatted_address
            assert r[1].formatted_address == expected_result[1].formatted_address


async def test_address_custom_api_call() -> None:
    with aioresponses() as m:
        raw_addrs = [raw_addr_obj2]
        expected_result = [std_addr_obj2]
        arrange_mongo()

        async def my_func(raw_addresses: List[RawAddress]) -> Dict[str, Any]:
            return response_data

        # act
        r: List[StandardizedAddress] = await AddressStandardizer().standardize_async(
            raw_addrs,
            cache_handler_obj=MockCacheHandler(),
            vendor_obj=cast(
                StandardizingVendor[BaseVendorApiResponse],
                MelissaStandardizingVendor(license_key="mock", custom_api_call=my_func),
            ),
        )

        # assert
        assert r[0].get_id() == "11"
        assert r[0].formatted_address == expected_result[0].formatted_address


async def test_documentdb_cache() -> None:
    # arrange
    arrange_mongo()
    raw_addrs = [raw_addr_obj, raw_addr_obj2]

    # act
    r = await DocumentDBCacheHandler().check_cache(raw_addrs)
    print(r.found)
    print(r.not_found)
    assert len(r.found) == 1
    assert len(r.not_found) == 1
    assert r.found[0].get_id() == "10"
    assert r.not_found[0].get_id() == "11"


def arrange_mongo() -> None:
    col: Collection[Any] = (
        pymongo.MongoClient(mongo_address)
        .get_database(mongo_collection_name)
        .get_collection(mongo_collection_name)
    )
    col.delete_many({})
    file_contents: List[Dict[str, Any]] = read_cache_file()
    cached_vendor_response: Dict[str, Any] = list(file_contents)[0]
    col.insert_one(cached_vendor_response)


async def test_standardize_stress_test() -> None:
    async def mocked_api_call(raw_addresses: List[RawAddress]) -> Dict[str, Any]:
        res = deepcopy(response_data)
        res["Records"] = [
            {
                "RecordID": a.address_id,
                "FormattedAddress": f"{a.to_str()}",
                "AddressLine1": a.line1,
                "AddressLine2": "",
                "Locality": a.city,
                "SubAdministrativeArea": "a county",
                "AdministrativeArea": a.state,
                "PostalCode": a.zipcode,
                "CountryISO3166_1_Alpha2": a.zipcode,
                "Latitude": "000",
                "Longitude": "-000",
            }
            for a in raw_addresses
        ]
        return res

    # arrange
    # random addresses
    address_population_count = 1000
    sample_count = address_population_count // 10
    random_address = []
    for i in range(address_population_count):
        random_id = random.randint(1, sample_count)
        random_address.append(
            {
                "id": str(random_id),
                "line1": f"line1 of {random_id}",
                "city": f"city{random_id}",
                "zip": f"zipcode {random_id}",
                "state": f"state {random_id}",
                "country": "US",
                "lat": f"lat{random_id}",
                "long": f"long{random_id}",
            }
        )
    # create raw address objects
    raw_addr_objs = [
        RawAddress(
            address_id=a["id"],
            line1=a["line1"],
            city=a["city"],
            state=a["state"],
            country=a["country"],
            zipcode=a["zip"],
        )
        for a in random_address
    ]
    # creat expected std response - for future use
    # std_addr_objs = [
    #     StdAddress(
    #         address_id=a["id"],
    #         line1=a["line1"],
    #         city=a["city"],
    #         state=a["state"],
    #         country=a["country"],
    #         zipcode=a["zip"],
    #         latitude=a["lat"],
    #         longitude=a["long"],
    #         formatted_address=f'a["line1"] a["city"] a["country"] a["zip"]',
    #     )
    #     for a in random_address
    # ]
    pymongo.MongoClient(mongo_address).get_database(
        mongo_collection_name
    ).get_collection(mongo_collection_name).delete_many({})

    # act
    r = await AddressStandardizer().standardize_async(
        raw_addresses=raw_addr_objs,
        vendor_obj=cast(
            StandardizingVendor[BaseVendorApiResponse],
            MelissaStandardizingVendor(
                license_key="mock", custom_api_call=mocked_api_call
            ),
        ),
        cache_handler_obj=DocumentDBCacheHandler(),
    )
    # assert
    assert len(r) == address_population_count
    assert (
        pymongo.MongoClient(mongo_address)
        .get_database(mongo_collection_name)
        .get_collection(mongo_collection_name)
        .count_documents({})
        == sample_count
    )


async def test_bad_request() -> None:
    async def mocked_api_call(raw_addresses: List[RawAddress]) -> Dict[str, Any]:
        res = deepcopy(response_data)
        res["Records"] = [
            {
                "RecordID": a.address_id,
                "FormattedAddress": "",
                "AddressLine1": "",
                "AddressLine2": "",
                "Locality": "",
                "SubAdministrativeArea": "",
                "AdministrativeArea": "",
                "PostalCode": "",
                "CountryISO3166_1_Alpha2": "US",
                "Latitude": "",
                "Longitude": "",
            }
            for a in raw_addresses
        ]
        return res

    # arrange
    # random addresses
    # create raw address objects
    empty_raw_address = RawAddress(
        address_id="1",
        line1="",
        city="",
        state="",
        country="US",
        zipcode="",
    )
    # creat expected std response

    pymongo.MongoClient(mongo_address).get_database(
        mongo_collection_name
    ).get_collection(mongo_collection_name).delete_many({})

    # act
    r = await AddressStandardizer().standardize_async(
        raw_addresses=[empty_raw_address],
        vendor_obj=cast(
            StandardizingVendor[BaseVendorApiResponse],
            MelissaStandardizingVendor(
                license_key="mock", custom_api_call=mocked_api_call
            ),
        ),
        cache_handler_obj=DocumentDBCacheHandler(),
    )
    # assert
    assert r[0].get_id() == "1"


async def test_none_cache() -> None:
    raw_addrs = [raw_addr_obj, raw_addr_obj2]
    r = await MockCacheHandler().check_cache(raw_addrs)
    assert len(r.not_found) == 2
    assert r.not_found[0].get_id() == "10"


async def test_vendor_http_error_call() -> None:
    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            status=500,
            payload={},
        )

        # arrange
        raw_addrs = [raw_addr_obj]

        # act / assert
        with pytest.raises(aiohttp.client_exceptions.ClientResponseError):
            await AddressStandardizer().standardize_async(
                raw_addrs,
                cache_handler_obj=MockCacheHandler(),
                vendor_obj=cast(
                    StandardizingVendor[BaseVendorApiResponse],
                    MelissaStandardizingVendor(
                        license_key="mock", response_key_error_threshold=0
                    ),
                ),
            )


async def test_vendor_empty_response_call() -> None:
    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            status=200,
            payload={},
        )

        # arrange
        raw_addrs = [raw_addr_obj]
        vendor = MelissaStandardizingVendor(
            license_key="mock", response_key_error_threshold=0
        )
        # act / assert
        with pytest.raises(VendorResponseKeyError):
            await AddressStandardizer().standardize_async(
                raw_addrs,
                cache_handler_obj=MockCacheHandler(),
                vendor_obj=cast(StandardizingVendor[BaseVendorApiResponse], vendor),
            )



import json
from dataclasses import dataclass, asdict
from typing import Optional, Dict, Any

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.api_helper.v2.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper


@dataclass
class OAuth2Credentails:
    client_id: str
    client_secret: str
    grant_type: Optional[str] = "client_credentials"
    audience: Optional[str] = None


class OAuth2ClientCredentialsFlow:
    def __init__(
        self,
        *,
        auth_url: str,
        auth_credentials: OAuth2Credentails,
        progress_logger: Optional[ProgressLogger],
        retry_count: int = 20,
        backoff_factor: float = 0.1,
        timeout_seconds: int = 120,
    ) -> None:
        """
        Encapsulate the logic to connect to OAuth2 servers using client credentials flow

        :param auth_url: OAuth2 token url
        :param auth_credentials: OAuth2 credentials like client id, secrets, etc
        :param progress_logger: Progress logger
        :param retry_count: Number of times to retry the request
        :param backoff_factor: Factor to backoff between retries
        :param timeout_seconds: Timeout in seconds
        """
        assert auth_url
        self.auth_url: str = auth_url
        self.auth_credentials = auth_credentials
        self.progress_logger: Optional[ProgressLogger] = progress_logger
        self.retry_count: int = retry_count
        self.backoff_factor: float = backoff_factor
        self.timeout_seconds: int = timeout_seconds

    def get_token(self) -> Optional[str]:
        http_request = HelixHttpRequest(
            url=self.auth_url,
            payload=asdict(
                self.auth_credentials,
                dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
            ),
            request_type=RequestType.POST,
            retry_count=self.retry_count,
            backoff_factor=self.backoff_factor,
            timeout_seconds=self.timeout_seconds,
        )

        response = AsyncHelper.run(http_request.get_result_async())
        if self.progress_logger:
            self.progress_logger.write_to_log(
                f"Received from {self.auth_url}: {json.dumps(response.result)}"
            )
        token = response.result.get("access_token")

        return token

    async def get_token_async(self) -> Optional[str]:
        http_request = HelixHttpRequest(
            url=self.auth_url,
            payload=asdict(
                self.auth_credentials,
                dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
            ),
            request_type=RequestType.POST,
            retry_count=self.retry_count,
            backoff_factor=self.backoff_factor,
            timeout_seconds=self.timeout_seconds,
        )

        response = await http_request.get_result_async()
        if self.progress_logger:
            self.progress_logger.write_to_log(
                f"Received from {self.auth_url}: {json.dumps(response.result)}"
            )
        token = response.result.get("access_token")
        return token

    async def get_token_response_async(self) -> Dict[str, Any]:
        http_request = HelixHttpRequest(
            url=self.auth_url,
            payload=asdict(
                self.auth_credentials,
                dict_factory=lambda x: {k: v for (k, v) in x if v is not None},
            ),
            request_type=RequestType.POST,
            retry_count=self.retry_count,
            backoff_factor=self.backoff_factor,
            timeout_seconds=self.timeout_seconds,
        )

        response = await http_request.get_result_async()
        if self.progress_logger:
            self.progress_logger.write_to_log(
                f"Received from {self.auth_url}: {json.dumps(response.result)}"
            )
        return response.result



import json
from os import path, makedirs
from pathlib import Path
from shutil import rmtree
from typing import Dict, Any, Union, cast

from mockserver_client.mock_requests_loader import load_mock_source_api_json_responses
from mockserver_client.mockserver_client import MockServerFriendlyClient
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import (
    StringType,
    StructField,
    StructType,
    Row,
    ArrayType,
)

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.http_data_sender.v2.http_data_sender import (
    HttpDataSender,
)
from spark_pipeline_framework.utilities.api_helper.http_request import (
    SingleJsonResult,
    SingleTextResult,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_http_data_sender(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_name = "test_http_data_sender"

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # noinspection PyTypeChecker
    spark_session.createDataFrame(  # type:ignore[type-var]
        [
            {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43919",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
            {
                "member_id": "e3a3f665-eae5-4046-a241-efdfe5c43910",
                "service_slug": "epic_local",
                "status": "Data Retrieved",
            },
        ],
    ).createOrReplaceTempView("my_view")

    server_url = f"http://mock-server:1080/{test_name}"
    mock_client = MockServerFriendlyClient("http://mock-server:1080")
    mock_client.clear(test_name)
    mock_client.reset()

    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("oauth_service"),
        mock_client=mock_client,
        url_prefix=f"{test_name}/token",
    )
    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("token_service"),
        mock_client=mock_client,
        url_prefix=test_name,
    )

    def payload_generator(json_data: Dict[str, Any]) -> Dict[str, Any]:
        temp_data = json_data.copy()
        temp_data.pop("status")
        return temp_data

    def response_processor(
        _: Dict[str, Any], response: Union[SingleJsonResult, SingleTextResult]
    ) -> str:
        return json.dumps([cast(Dict[str, Any], response.result)])

    # Act
    with ProgressLogger() as progress_logger:
        HttpDataSender(
            progress_logger=progress_logger,
            source_view="my_view",
            view="output_view",
            url=f"{server_url}",
            batch_count=1,
            client_id="client_id",
            client_secret="client_secret",
            auth_url=f"{server_url}/token",
            payload_generator=payload_generator,
            response_processor=response_processor,
            parse_response_as_json=True,
            response_schema=ArrayType(
                StructType(
                    [
                        StructField("access_token", StringType(), True),
                        StructField("expires_in", StringType(), True),
                        StructField("token_type", StringType(), True),
                    ]
                )
            ),
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table("output_view")
    result_df.printSchema()
    result_df.show(truncate=False)

    assert result_df.collect()[0]["result"] == [
        Row(access_token="fake access_token", expires_in="54000", token_type="bearer")
    ]
    assert result_df.collect()[1]["result"] == [
        Row(access_token="fake access_token2", expires_in="54000", token_type="bearer")
    ]



import json
from typing import Any, Callable, Dict, List, Optional, Union

from spark_pipeline_framework.utilities.api_helper.http_request import HelixHttpRequest
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v1.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import sc


class HttpDataReceiver(FrameworkTransformer):
    """
    This is a generic class to call a http api and return the response
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        http_request: HelixHttpRequest,
        view_name: str,
        name: Optional[str] = None,
        one_iteration_only: Optional[bool] = False,
        next_request_generator: Optional[
            Callable[
                [HelixHttpRequest, Dict[str, Any], Optional[ProgressLogger]],
                Optional[HelixHttpRequest],
            ]
        ] = None,
        response_processor: Optional[
            Callable[
                [
                    List[Dict[str, Any]],
                    Union[List[Dict[str, Any]], Dict[str, Any]],
                    Optional[ProgressLogger],
                ],
                List[Dict[str, Any]],
            ]
        ] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        log_response: bool = False,
        auth_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ) -> None:
        """
        Transformer to call and receive data from an API


        :param http_request: HelixHttpRequest to specify the url and type of the request
        :param view_name: name of the view to read the response into
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        :param next_request_generator: implement this function to keep calling the API and adding to response list.
        it supposed to return a HelixHttpRequest to be used to call the API or return None to end the API call loop
        :param response_processor: it can change the result before loading to spark df
        :param auth_url: (Optional) url to use to authenticate with client credentials
        :param client_id: (Optional) client id to use to authenticate with client credentials
        :param client_secret: (Optional) client secret to use to authenticate with client credentials
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.http_request: Param[HelixHttpRequest] = Param(self, "http_request", "")
        self._setDefault(http_request=None)

        self.view_name: Param[str] = Param(self, "view_name", "")
        self._setDefault(view_name=None)

        self.one_iteration_only: Param[str] = Param(self, "one_iteration_only", "")
        self._setDefault(one_iteration_only=False)

        self.log_response: Param[bool] = Param(self, "log_response", "")
        self._setDefault(log_response=log_response)

        self.next_request_generator: Param[
            Optional[
                Callable[
                    [HelixHttpRequest, Dict[str, Any], Optional[ProgressLogger]],
                    Optional[HelixHttpRequest],
                ]
            ]
        ] = Param(self, "next_request_generator", "")
        self._setDefault(next_request_generator=None)

        self.response_processor: Param[
            Optional[
                Callable[
                    [List[Dict[str, Any]], Dict[str, Any], Optional[ProgressLogger]],
                    List[Dict[str, Any]],
                ]
            ]
        ] = Param(self, "response_processor", "")
        self._setDefault(response_processor=None)

        self.auth_url: Param[Optional[str]] = Param(self, "auth_url", "")
        self._setDefault(auth_url=None)

        self.client_id: Param[Optional[str]] = Param(self, "client_id", "")
        self._setDefault(client_id=None)

        self.client_secret: Param[Optional[str]] = Param(self, "client_secret", "")
        self._setDefault(client_secret=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        http_request: Optional[HelixHttpRequest] = self.getHttpRequest()
        view_name = self.getView()
        one_iteration_only = self.getOneIteration()
        name: Optional[str] = self.getName()

        auth_url: Optional[str] = self.getOrDefault(self.auth_url)
        client_id: Optional[str] = self.getOrDefault(self.client_id)
        client_secret: Optional[str] = self.getOrDefault(self.client_secret)

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=f"{name}_http_data_receiver", progress_logger=progress_logger
        ):
            if client_id and auth_url and client_secret and http_request:
                # first call auth to get a token
                oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
                    OAuth2ClientCredentialsFlow(
                        auth_url=auth_url,
                        client_id=client_id,
                        client_secret=client_secret,
                        progress_logger=progress_logger,
                    )
                )

                access_token: Optional[str] = oauth2_client_credentials_flow.get_token()

                if access_token:
                    if not http_request.headers:
                        http_request.headers = {}
                    http_request.headers["Authorization"] = f"Bearer {access_token}"

            responses: List[Dict[str, Any]] = []
            try:
                while True:
                    assert http_request is not None
                    self.logger.debug(f"Calling API: {http_request.to_string()}...")
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Calling API: {http_request.to_string()}..."
                        )
                    response = http_request.get_result()

                    self.logger.info(
                        f"Successfully retrieved: {http_request.url} with status {response.status}"
                    )
                    if self.getLogResponse():
                        self.logger.info(
                            f"Response: {json.dumps(response.result, default=str)}"
                        )
                    if progress_logger:
                        progress_logger.write_to_log(
                            f"Successfully retrieved: {http_request.url} with status {response.status}"
                        )
                        if self.getLogResponse():
                            progress_logger.write_to_log(
                                f"Response [{response.status}]: {json.dumps(response.result, default=str)}"
                            )

                    next_request_generator = self.getNextRequestGenerator()
                    http_request = (
                        next_request_generator(
                            http_request, response.result, progress_logger
                        )
                        if next_request_generator and not one_iteration_only
                        else None
                    )
                    # accumulated responses before loading to spark
                    response_processor = self.getResponseProcessor()
                    if response_processor:
                        responses = response_processor(
                            responses, response.result, progress_logger
                        )
                    else:
                        responses.append(response.result)

                    if http_request is None or not response.result:
                        break

                df2 = df.sparkSession.read.json(
                    sc(df).parallelize([json.dumps(r) for r in responses])
                )
                df2.createOrReplaceTempView(view_name)
            except Exception as e:
                url = (
                    http_request.url
                    if (http_request and hasattr(http_request, "url"))
                    else None
                )
                error_text: str = f"Url: {url}: {http_request} {repr(e)}"
                self.logger.error(error_text)
                raise Exception(error_text) from e

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHttpRequest(self) -> HelixHttpRequest:
        return self.getOrDefault(self.http_request)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOneIteration(self) -> str:
        return self.getOrDefault(self.one_iteration_only)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getNextRequestGenerator(
        self,
    ) -> Optional[
        Callable[
            [HelixHttpRequest, Dict[str, Any], Optional[ProgressLogger]],
            Optional[HelixHttpRequest],
        ]
    ]:
        return self.getOrDefault(self.next_request_generator)

    # noinspection PyPep8Naming
    def getResponseProcessor(
        self,
    ) -> Optional[
        Callable[
            [List[Dict[str, Any]], Dict[str, Any], Optional[ProgressLogger]],
            List[Dict[str, Any]],
        ]
    ]:
        return self.getOrDefault(self.response_processor)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLogResponse(self) -> bool:
        return self.getOrDefault(self.log_response)



from typing import Dict, Any, Optional

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.pyspark_helpers.clean_column_name import (
    replace_special_characters_in_columns_in_data_frame,
)


class FrameworkColumnNameCleaner(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        name: Optional[str] = None,
        view: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        This transfromer replaces any column names with invalid characters with _


        :param name: name of transformer
        :param view: (Optional) view to read from and update
        :param parameters:
        :param progress_logger:
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        # name: Optional[str] = self.getName()

        if view:
            df = df.sparkSession.table(view)

        df = replace_special_characters_in_columns_in_data_frame(df=df)

        if view:
            df.createOrReplaceTempView(view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)



from typing import Dict, Optional


from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)


class StandardizedAddress(RawAddress):
    """
    The address standardized by a vendor
    """

    county: Optional[str] = None
    longitude: Optional[str] = None
    latitude: Optional[str] = None
    formatted_address: Optional[str] = None
    standardize_vendor: str

    def to_str(self) -> str:
        if self.formatted_address:
            return str(self.formatted_address)
        else:
            return super(StandardizedAddress, self).to_str()

    def to_dict(self) -> Dict[str, str]:
        # noinspection PyProtectedMember
        return self.model_dump()

    @classmethod
    def from_dict(cls, address_dict: Dict[str, str]) -> "StandardizedAddress":
        return cls(**address_dict)

    @classmethod
    def from_raw_address(
        cls, raw_address: RawAddress, vendor_name: str
    ) -> "StandardizedAddress":
        assert vendor_name, "vendor_name is required"
        return cls(**raw_address.to_dict(), standardize_vendor=vendor_name)

    @classmethod
    def from_address_id(
        cls, address_id: str, vendor_name: str
    ) -> "StandardizedAddress":
        assert address_id, "address_id is required"
        assert vendor_name, "vendor_name is required"
        return cls(
            address_id=address_id,
            standardize_vendor=vendor_name,
            line1=None,
            line2=None,
            city=None,
            state=None,
            zipcode=None,
        )

    def is_valid(self) -> bool:
        return (
            self.zipcode is not None
            and self.city is not None
            and self.state is not None
        )



import json
from os import path, makedirs, environ
from pathlib import Path
from shutil import rmtree
from urllib.parse import urljoin

import pytest
from helix_fhir_client_sdk.responses.fhir_delete_response import FhirDeleteResponse
from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.fhir_sender.v2.fhir_sender import FhirSender
from spark_pipeline_framework.utilities.fhir_server_test_context.v1.fhir_server_test_context import (
    FhirServerTestContext,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

import requests


@pytest.mark.parametrize("run_synchronously", [True, False])
async def test_fhir_sender_merge_parquet(
    spark_session: SparkSession, run_synchronously: bool
) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    test_files_dir: Path = data_dir.joinpath("test_files/patients")  # json

    async with FhirServerTestContext(
        resource_type="Patient"
    ) as fhir_server_test_context:

        # convert to parquet
        test_files_dir_parquet: Path = data_dir.joinpath("test_files/patients_parquet")
        df: DataFrame = spark_session.read.format("json").load(str(test_files_dir))
        df.write.mode("overwrite").format("parquet").save(str(test_files_dir_parquet))

        df = spark_session.read.format("parquet").load(str(test_files_dir_parquet))

        response_files_dir: Path = temp_folder.joinpath("patients-response")

        df = create_empty_dataframe(spark_session=spark_session)

        fhir_server_url: str = fhir_server_test_context.fhir_server_url
        auth_client_id = fhir_server_test_context.auth_client_id
        auth_client_secret = fhir_server_test_context.auth_client_secret
        auth_well_known_url = fhir_server_test_context.auth_well_known_url

        # first delete any existing resources
        fhir_client = await fhir_server_test_context.create_fhir_client_async()

        fhir_client = fhir_client.url(fhir_server_url).resource("Patient")
        delete_response: FhirDeleteResponse = await fhir_client.id_(
            "00100000000"
        ).delete_async()
        assert delete_response.status == 204
        delete_response = await fhir_client.id_("00200000000").delete_async()
        assert delete_response.status == 204

        token_url = fhir_server_test_context.get_token_url()
        assert token_url
        authorization_header = fhir_server_test_context.get_authorization_header()

        environ["LOGLEVEL"] = "DEBUG"
        # Act
        with ProgressLogger() as progress_logger:
            await FhirSender(
                resource="Patient",
                server_url=fhir_server_url,
                file_path=test_files_dir_parquet,
                response_path=response_files_dir,
                progress_logger=progress_logger,
                batch_size=1,
                file_format="parquet",
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_well_known_url=auth_well_known_url,
                run_synchronously=run_synchronously,
            ).transform_async(df)

        # Assert
        response = requests.get(
            urljoin(fhir_server_url, "Patient/00100000000"),
            headers=authorization_header,
        )
        assert response.ok, response.text
        json_text: str = response.text
        obj = json.loads(json_text)
        assert obj["birthDate"] == "2017-01-01"

        response = requests.get(
            urljoin(fhir_server_url, "Patient/00200000000"),
            headers=authorization_header,
        )
        assert response.ok
        json_text = response.text
        obj = json.loads(json_text)
        assert obj["birthDate"] == "1984-01-01"



from typing import Dict, Any, Callable, Union, List, cast

from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from spark_pipeline_framework.pipelines.v2.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_transformer_helpers import (
    get_python_function_from_location,
)
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class MappingPipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(MappingPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="987654321",
            client_name="client_bar",
            vendor_name="vendor_bar",
        )
        mapping_function: Callable[
            [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
        ] = get_python_function_from_location(
            location=str(parameters["feature_path"]),
            import_module_name=".mapping",
            function_name="mapping",
        )

        self.transformers = self.create_steps(
            cast(
                List[FrameworkTransformer],
                [
                    FrameworkMappingLoader(
                        view="members",
                        mapping_function=mapping_function,
                        parameters=parameters,
                        progress_logger=progress_logger,
                    )
                ],
            )
        )



import hashlib
import json
import uuid
from typing import Dict, Optional

from pydantic import BaseModel, Field


class RawAddress(BaseModel):
    """
    The address that needs to get standardized
    """

    address_id: str
    line1: Optional[str]
    line2: Optional[str] = None
    city: Optional[str]
    state: Optional[str]
    zipcode: Optional[str]
    country: Optional[str] = "US"

    internal_id: str = Field(
        default_factory=lambda: str(uuid.uuid4())
    )  # Generate a random internal id

    def to_dict(self) -> Dict[str, str]:
        return self.model_dump()

    @classmethod
    def from_dict(cls, address_dict: Dict[str, str]) -> "RawAddress":
        return cls(**address_dict)

    def get_id(self) -> str:
        address_id: str = self.address_id
        return address_id

    def get_internal_id(self) -> str:
        """Get the internal id of the address"""
        return str(self.internal_id)

    def set_id(self, address_id: str) -> None:
        self.address_id = address_id

    def to_str(self) -> str:
        # todo: make sure the online format is legit
        line2 = " " + self.line2 if self.line2 else ""
        addr: str = (
            f"{self.line1 or ''}{line2}, {self.city or ''} {self.state or ''} {self.zipcode or ''} {self.country or ''}"
        )
        return addr.replace("  ", " ").strip(" ,")

    def to_hash(self) -> str:
        # reducing variance by cleaning the address
        a_dict: Dict[str, str] = {
            k: str(v if v else "").strip().lower() for k, v in self.to_dict().items()
        }
        addr: str = (
            f'{a_dict["line1"]}{a_dict["line2"]}{a_dict["city"]}{a_dict["state"]}{a_dict["zipcode"]}{a_dict["country"]}'
        )
        # create hash
        return hashlib.sha1(addr.encode()).hexdigest()

    # noinspection PyMethodMayBeStatic
    def _check_id_unique(self) -> bool:
        # todo check if address_id is unique
        return False

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def is_valid_for_geolocation(self) -> bool:
        """
        Check if the address is valid for geolocation
        """
        return all([self.line1, self.city, self.state, self.zipcode, self.country])



from setuptools import setup, find_packages
from os import path, getcwd

# from https://packaging.python.org/tutorials/packaging-projects/

# noinspection SpellCheckingInspection
package_name = "sparkpipelineframework"

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    with open(path.join(getcwd(), "VERSION")) as version_file:
        version = version_file.read().strip()
except IOError:
    raise


# classifiers list is here: https://pypi.org/classifiers/

# create the package setup
setup(
    name=package_name,
    version=version,
    author="Imran Qureshi",
    author_email="imranq2@hotmail.com",
    description="Framework for simpler Spark Pipelines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/icanbwell/SparkPipelineFramework",
    packages=find_packages(
        exclude=[
            "*/test",
            "*/test/*",
            "*tests/*",
            "*library/*",
            "*library",
            "docs",
            "docs/*",
            "docsrc",
            "docsrc/*",
            "keycloak-config",
            "keycloak-config/*",
            ".gihub",
            ".github/*",
        ]
    ),
    install_requires=[
        "protobuf>=3",
        "pyspark==3.5.1",
        "pyarrow>=17.0.0",
        "delta-spark==3.2.0",
        "sparkautomapper>=3.0.1",
        "pymysql>=1.1.1",
        "furl>=2.1.3",
        "requests>=2.32.3",
        "boto3>=1.34.140",
        "chardet",
        "slack-sdk>=3.22.0",
        "smart-open[s3]>=6.3.0",
        "sqlalchemy>=1.4.37",
        "sqlparse>=0.5.3",
        "bounded-pool-executor>=0.0.3",
        "fastjsonschema>=2.18.0",
        "helix.fhir.client.sdk>=4.1.11",
        "opensearch-py[async]>=2.6.0",
        "pyathena>2.14.0",
        "spark-nlp>=4.2.3",
        "pymongo>=4.10.1",
        "pandas>=2",
        "structlog>=23.1.0",
        "usaddress>=0.5.10",
        "usaddress-scourgify>=0.6.0",
        "aiohttp>=3.10.11",
        "pydantic>=2.8.2",
        "motor[snappy,zstd]>=3.5.1",
        "dataclasses-json>=0.6.7",
        "helixtelemetry>=1.0.3",
        "opentelemetry-api>=1.30.0",
        "opentelemetry-sdk>=1.30.0",
        "opentelemetry-exporter-otlp>=1.30.0",
        "opentelemetry-instrumentation-aiohttp-client>=0.51b0",
        "msgpack>=1.1.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    dependency_links=[],
    include_package_data=True,
    zip_safe=False,
    package_data={"spark_pipeline_framework": ["py.typed"]},
)



from pathlib import Path
from typing import Dict, Any, Union, Optional, Callable

from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.errors import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
    create_empty_dataframe,
)


class FrameworkJsonExporter(FrameworkTransformer):
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: Optional[int] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        throw_if_empty: bool = False,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ) -> None:
        """
        Exports the data frame to JSON


        :param view: view to load from parquet
        :param file_path: where to load from
        :param name: a name for the transformer step
        :param parameters: parameters
        :param progress_logger: the logger to use for logging
        :param throw_if_empty: throw an error if the passed in data frame is empty
        :param stream: whether to stream the data
        :param delta_lake_table: store data in this delta lake
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert mode in FileWriteModes.MODE_CHOICES

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=limit)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.throw_if_empty: Param[bool] = Param(self, "throw_if_empty", "")
        self._setDefault(throw_if_empty=throw_if_empty)

        self.stream: Param[bool] = Param(self, "stream", "")
        self._setDefault(stream=stream)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=delta_lake_table)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        throw_if_empty: bool = self.getThrowIfEmpty()
        stream: bool = self.getStream()
        limit: Optional[int] = self.getLimit()
        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        file_format: str = "delta" if delta_lake_table else "json"

        with ProgressLogMetric(
            name=f"{name or view}_fhir_exporter", progress_logger=progress_logger
        ):
            try:
                mode = self.getMode()
                if view:
                    df_view: DataFrame = df.sparkSession.table(view)
                    assert not throw_if_empty or not spark_is_data_frame_empty(
                        df=df_view
                    )
                    if spark_is_data_frame_empty(df=df_view):
                        df_view = create_empty_dataframe(df.sparkSession)
                    if limit is not None and limit >= 0:
                        df_view = df_view.limit(limit)
                    if stream:
                        df_view.writeStream.format(file_format).option(
                            "path", str(file_path)
                        ).start()
                    else:
                        df_view.write.format(file_format).mode(mode).save(
                            path=str(file_path)
                        )
                else:
                    assert not throw_if_empty or not spark_is_data_frame_empty(df=df)
                    if not spark_is_data_frame_empty(df=df):
                        df = create_empty_dataframe(df.sparkSession)
                    if limit is not None and limit >= 0:
                        df = df.limit(limit)
                    if stream:
                        df.writeStream.format(file_format).option(
                            "path", str(file_path)
                        ).start()
                    else:
                        df.write.format(file_format).mode(mode).save(
                            path=str(file_path)
                        )

                if progress_logger:
                    progress_logger.log_param("data_export_path", str(file_path))
                    progress_logger.write_to_log(
                        f"[{name or view}] written to {file_path}"
                    )

            except AnalysisException as e:
                self.logger.error(f"[{name or view}] File write failed to {file_path}")
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getThrowIfEmpty(self) -> bool:
        return self.getOrDefault(self.throw_if_empty)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getStream(self) -> bool:
        return self.getOrDefault(self.stream)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDeltaLakeTable(self) -> Optional[str]:
        return self.getOrDefault(self.delta_lake_table)
